import time
import os
import psutil
while True:
    for p in psutil.process_iter():
        if 'sleep' in p.cmdline():
            print('pid', p.pid, p.cmdline(), 'sid', os.getpgid(p.pid))
    time.sleep(.1)



from pathlib import Path
import json
from loguru import logger
import pytest
import pypipegraph2 as ppg
from .shared import write, read, counter
import os


def get_known(job):
    known = {}
    for fn in (job.output_dir_prefix / "used_by").glob("*.source"):
        key = fn.read_text().strip()
        value = Path(os.readlink(fn.with_suffix(".uses"))).name
        known[key] = value
    return known


@pytest.mark.usefixtures("ppg2_per_test")
class TestSharedJob:
    def test_simple(self, job_trace_log):
        def doit(output_files, output_prefix):
            for of in output_files:
                assert not "no_input" in str(of)
            count = counter("doit")
            write(output_files[0], "a" + str(count))
            write(output_files[1], "b")

        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False, remove_unused=False
        )

        def followup_c(of):
            of.write_text(read(job.files[0]) + read(job.files[1]))
            counter("c")

        def followup_d(of):
            of.write_text(read(job.files[1]))
            counter("d")

        ppg.FileGeneratingJob("C", followup_c).depends_on(job)
        ppg.FileGeneratingJob("D", followup_d).depends_on(job.files[1])

        ppg.util.log_error("one")
        ppg.run()
        assert read(job.find_file("a")) == "a0"
        assert read(job.find_file("b")) == "b"
        assert read("doit") == "1"
        assert read("C") == "a0b"
        assert read("D") == "b"
        assert read("c") == "1"
        assert read("d") == "1"

        ppg.util.log_error("two")
        ppg.run()
        assert read("doit") == "1"
        assert read("c") == "1"
        assert read("d") == "1"

        job.depends_on(ppg.ParameterInvariant("E", "e"))
        ppg.util.log_error("three")
        ppg.run()
        assert read("doit") == "2"
        assert read("c") == "2"
        assert read("d") == "1"
        assert read(job.files[0]) == "a1"
        assert read(job.files[1]) == "b"
        assert read("C") == "a1b"

        ppg.new(log_level=6)
        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False, remove_unused=False
        )
        ppg.FileGeneratingJob("C", followup_c).depends_on(job)
        ppg.FileGeneratingJob("D", followup_d).depends_on(job.files[1])

        job.depends_on(ppg.ParameterInvariant("E", "f"))
        ppg.util.log_error("three")
        ppg.run()
        assert read("doit") == "3"
        assert read("c") == "3"
        assert read("d") == "1"
        assert read(job.files[0]) == "a2"
        assert read(job.files[1]) == "b"
        assert read("C") == "a2b"
        assert len(list(job.output_dir.glob("*"))) == 3  # three different outputs
        assert len(list(job.input_dir.glob("*"))) == 3  # three different outputs

        job.remove_unused = True
        ppg.util.log_error("four")
        assert read("doit") == "3"
        ppg.run()
        assert read("doit") == "3"

        # no build, have history -> no cleanup
        assert len(list(job.output_dir.glob("*"))) == 3  # three different outputs
        assert len(list(job.input_dir.glob("*"))) == 3  # three different outputs
        ppg.global_pipegraph.get_history_filename().unlink()
        ppg.run()
        assert (
            len(list(job.output_dir.glob("*"))) == 1
        )  # loosing history is reason for cleanup
        assert (
            len(list(job.input_dir.glob("*"))) == 1
        )  # loosing history is reason for cleanup

    def test_multi_file_gen_job_lookup_colission(self):
        with pytest.raises(ValueError):
            ppg.MultiFileGeneratingJob({"a": "A", "b": "A"}, lambda of: None)
        with pytest.raises(ValueError):
            ppg.SharedMultiFileGeneratingJob(
                "shared", {"a": "A", "b": "A"}, lambda of, prefix: None
            )

    def test_subdirs(self):
        def doit(files, prefix):
            for f in files.values():
                f.parent.mkdir()
                f.write_text(f.name)

        job = ppg.SharedMultiFileGeneratingJob(
            "out", {"a": "a/a", "b": "b/b"}, doit, depend_on_function=False
        )
        with pytest.raises(AttributeError):
            job["a"]  # can't get a filename *before* the job has run
        # assert "__never_placed_here__" in str(job["a"])
        # job()
        ppg.run()
        assert read(job["a"]) == "a"
        assert read(job["b"]) == "b"

    def test_nested(self):
        def doit(output_files, prefix):
            count = counter("doit")
            write(output_files[0], "a" + str(count))
            write(output_files[1], "b")

        def func_c(files, prefix):
            counter("C")
            files[0].write_text(read(job[0]) + "c")

        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False
        )
        c = ppg.SharedMultiFileGeneratingJob("out2", ["c"], func_c)
        c.depends_on(job)
        ppg.FileGeneratingJob(
            "d", lambda of: counter("d") and of.write_text(read(c[0]) + "d")
        ).depends_on(c)
        ppg.run()

        assert read(job.find_file("a")) == "a0"
        assert read(job.find_file("b")) == "b"
        assert read(c[0]) == "a0c"
        assert read("C") == "1"
        assert read("d") == "a0cd"

        ppg.new()
        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False
        )
        c = ppg.SharedMultiFileGeneratingJob("out2", ["c"], func_c)
        c.depends_on(job)
        d = ppg.FileGeneratingJob(
            "d", lambda of: counter("d") and of.write_text(read(c[0]) + "d")
        ).depends_on(c)
        ppg.run()
        assert read(job.find_file("a")) == "a0"
        assert read(job.find_file("b")) == "b"

        assert read(c[0]) == "a0c"
        assert read("C") == "1"
        assert read("d") == "a0cd"

        ppg.global_pipegraph.get_history_filename().unlink()
        ppg.new()

        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False
        )
        c = ppg.SharedMultiFileGeneratingJob("out2", ["c"], func_c)
        c.depends_on(job)
        ppg.FileGeneratingJob(
            "d", lambda of: counter("d") and of.write_text(read(c[0]) + "d")
        ).depends_on(c)
        # subprocess.check_call(["fd", "-L"])
        ppg.util.log_error("last before bookm")
        ppg.run()

        assert read(job.find_file("a")) == "a0"
        assert read(job.find_file("b")) == "b"

        assert read(c[0]) == "a0c"
        assert read("C") == "1"
        assert read("d") == "a0cd"

    def test_nuking_on_error(self):
        def doit(output_files, prefix):
            raise ValueError()

        a = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit)
        with pytest.raises(ppg.JobsFailed):
            a()
        assert not list(Path("out/build").glob("*"))
        a = ppg.SharedMultiFileGeneratingJob(
            "out", ["a"], doit, remove_build_dir_on_error=False
        )
        with pytest.raises(ppg.JobsFailed):
            a()
        assert list(Path("out/build").glob("*"))

    def test_multiple_histories(self):
        def doit(output_files, prefix):
            for of in output_files:
                assert not "no_input" in str(of)
            count = counter("doit")
            write(output_files[0], "a" + str(count))
            write(output_files[1], "b")

        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False, remove_unused=False
        )
        ppg.FileGeneratingJob(
            "a",
            lambda of: counter("A") and of.write_text(str(of)),
            depend_on_function=False,
        ).depends_on(job)
        ppg.run()
        assert read("A") == "1"
        h1 = ppg.global_pipegraph.get_history_filename()
        assert read("doit") == "1"

        ppg.new(history_dir=".my_history")
        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False, remove_unused=False
        )
        ppg.FileGeneratingJob(
            "a",
            lambda of: counter("A") and of.write_text(str(of)),
            depend_on_function=False,
        ).depends_on(job)
        h2 = ppg.global_pipegraph.get_history_filename()
        assert h1.resolve() != h2.resolve()
        logger.info("2nd run")
        ppg.run()
        assert read("doit") == "1"
        assert (
            read("A") == "2"
        )  # changing the history dir obviously triggers a rerun to capture hashes.

        ppg.new(history_dir=".my_history2")
        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False, remove_unused=False
        )
        job.depends_on(ppg.ParameterInvariant("e", "e"))
        ppg.FileGeneratingJob(
            "a",
            lambda of: counter("A") and of.write_text(str(of)),
            depend_on_function=False,
        ).depends_on(job)
        h3 = ppg.global_pipegraph.get_history_filename()
        assert h3.resolve() != h2.resolve()
        ppg.run()
        assert read("doit") == "2"
        assert read("A") == "3"  # changed history dir...

        known = get_known(job)
        ha2 = ppg.SharedMultiFileGeneratingJob._handle_anysnake2

        assert ha2(str(h1.absolute())) in known
        assert ha2(str(h2.absolute())) in known
        assert ha2(str(h3.absolute())) in known
        assert known[ha2(str(h1.absolute()))] == known[ha2(str(h2.absolute()))]
        assert known[ha2(str(h1.absolute()))] != known[ha2(str(h3.absolute()))]
        key3 = known[ha2(str(h3.absolute()))]
        assert (Path("out/by_input") / key3).exists()

        ppg.run()  # running again is harmless
        assert read("doit") == "2"
        assert read("A") == "3"

        logger.info("final")
        ppg.new(history_dir=".my_history2")
        job = ppg.SharedMultiFileGeneratingJob(
            "out", ["a", "b"], doit, depend_on_function=False, remove_unused=False
        )
        ppg.FileGeneratingJob(
            "a",
            lambda of: counter("A") and of.write_text(str(of)),
            depend_on_function=False,
        ).depends_on(job)
        h4 = ppg.global_pipegraph.get_history_filename()
        assert h3.resolve() == h4.resolve()
        ppg.run()
        assert read("A") == "4"  # changed input
        known = get_known(job)
        assert ha2(str(h3.absolute())) in known
        print(job.target_folder)
        print(known)
        assert (
            known[ha2(str(h3.absolute()))] == known[ha2(str(h2.absolute()))]
        )  # goes back to old value
        assert not (Path("out") / key3).exists()

    def test_simple_one_file(self):
        def doit(output_file, prefix):
            count = counter("doit")
            write(output_file[0], "a" + str(count))

        job = ppg.SharedMultiFileGeneratingJob(
            "out",
            ["a"],
            doit,
            depend_on_function=False,
            remove_unused=False,
            remove_build_dir_on_error=False,
        )
        with pytest.raises(AttributeError):
            job.target_folder
        try:
            ppg.run()
        except Exception:
            print("stdout", job.stdout)
            print("stderr", job.stderr)
            raise
        assert read("out/by_input/no_input/a") == "a0"
        assert read("doit") == "1"
        ppg.run()
        assert read("out/by_input/no_input/a") == "a0"
        assert read("doit") == "1"
        before = job.target_folder
        job.depends_on(ppg.ParameterInvariant("B", "shu"))
        with pytest.raises(AttributeError):
            job.target_folder
        ppg.run()
        after = job.target_folder
        assert before != after

    def test_remove_and_keep_build_dir(self):
        def dofail(ofs, prefix):
            raise ValueError()

        jobKeep = ppg.SharedMultiFileGeneratingJob(
            "out",
            ["a"],
            dofail,
            depend_on_function=False,
            remove_unused=False,
            remove_build_dir_on_error=False,
        )
        jobTrash = ppg.SharedMultiFileGeneratingJob(
            "outB",
            ["a"],
            dofail,
            depend_on_function=False,
            remove_unused=False,
            remove_build_dir_on_error=True,
        )
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert jobKeep.target_folder.exists()
        assert not jobTrash.target_folder.exists()
        assert "/build/" in str(jobKeep.target_folder)

    def test_direct_call(self):
        def doit(output_file, prefix):
            count = counter("doit")
            write(output_file[0], "a" + str(count))

        job = ppg.SharedMultiFileGeneratingJob(
            "out",
            ["a"],
            doit,
            depend_on_function=False,
            remove_unused=False,
            remove_build_dir_on_error=False,
        )
        job()
        assert (job.target_folder / "a").read_text() == "a0"

    def test_shared_job_with_dict_file_def(self):
        def doit(output_files, prefix):
            counter("doit")
            write(output_files["a"], "file_a")

        job = ppg.SharedMultiFileGeneratingJob("out", {"a": "file_a"}, doit)
        ppg.run()
        assert read(job.find_file("a")) == "file_a"
        # assert read(job.find_file("file_a")) == "file_a"
        assert read("doit") == "1"

        ppg.new(history_dir=".my_history")
        job = ppg.SharedMultiFileGeneratingJob("out", {"b": "file_a"}, doit)
        ppg.run()
        assert read(job.find_file("b")) == "file_a"
        # assert read(job.find_file("file_a")) == "file_a"
        assert read(job["b"]) == "file_a"
        # assert read(job["file_a"]) == "file_a"
        with pytest.raises(KeyError):
            job["file_a"]
        with pytest.raises(KeyError):
            job["c"]
        assert read("doit") == "1"
        res = job()
        assert set(res.keys()) == {"b"}
        assert read(res["b"]) == "file_a"

    def test_shared_job_with_changing_inputs(self):
        ppg.new(history_dir="history_one")
        out = ["a"]

        def doit(output_files, prefix, out=out):
            counter("doit")
            output_files[0].write_text("A" + out[0])

        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit, remove_unused=False)
        ppg.run()
        assert read("doit") == "1"
        assert read(job["a"]) == "Aa"
        with pytest.raises(KeyError):
            job["b"]
        job.depends_on_params("shu")
        ppg.run()
        assert read("doit") == "2"
        assert len(list(Path("out/done").glob("*"))) == 1
        assert read(job["a"]) == "Aa"
        out[0] = "b"
        known = get_known(job)
        assert len(known) == 1

        ppg.new(history_dir="my_history")
        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit, remove_unused=False)
        job.depends_on_params("sha")
        job()
        assert read("doit") == "3"
        assert len(list(Path("out/done").glob("*"))) == 2
        assert read(job["a"]) == "Ab"

        # we had run the job, we had hashes, the output files do exist
        # thanks to running in another history,
        # so the job reruns, but doesn't recalc
        # and we get no new output dir
        # and we finally trigger the last case in .output_needed

        ppg.new(history_dir="history_one")
        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit, remove_unused=False)
        job.depends_on_params("sha")
        job()
        assert read("doit") == "3"  # we do not rerun the code..
        assert len(list(Path("out/done").glob("*"))) == 2
        assert read(job["a"]) == "Ab"
        new_known = get_known(job)
        k = list(known.keys())[0]
        assert new_known[k] != known[k]
        assert len(new_known) == 2

    def test_violation(self):
        """Test that the detecting of non-detemernistic outputs,
        best efforts as it is,
        will at least capture the case where two builds run in parallel
        and differ in outpu
        """
        import time

        symlink = []

        def doit(output_files, prefix):
            r = counter("doit")
            output_files[0].write_text(str(time.time()))
            time.sleep(0.1)
            if (
                r == "1"
            ):  # sneakily restore the symlink. So we can trigger the violation case
                symlink[0].symlink_to(symlink[1])

        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit)
        ppg.run()
        assert read("doit") == "1"

        # now take away the symlink so the alterante history has to run again
        sl = list(Path("out/by_input").glob("*"))
        assert len(sl) == 1
        symlink.append(sl[0])
        ol = list(Path("out/done").glob("*"))
        assert len(ol) == 1
        symlink.append(ol[0])
        os.unlink(sl[0])
        ppg.new(history_dir="history_two")
        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit, remove_unused=False)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert "two different outputs" in str(job.exception)
        assert read("doit") == "2"

    def test_some_files_missing(self):
        def doit(output_files, prefix):
            for f in output_files:
                f.write_text(f.name)

        job = ppg.SharedMultiFileGeneratingJob("out", ["a", "b"], doit)
        ppg.run()
        job["b"].unlink()
        # with pytest.raises(ppg.JobContractError) as excinfo:
        # ppg.run()
        # assert 'some result files' in str(excinfo.value)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert "some result files" in str(job.exception)

        # second path to that exception
        ppg.new(history_dir="history2")
        job = ppg.SharedMultiFileGeneratingJob("out", ["a", "b"], doit)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert "some result files" in str(job.exception)

    def test_non_symlink_in_folder(self):
        def doit(output_files, prefix):
            for f in output_files:
                f.write_text(f.name)

        job = ppg.SharedMultiFileGeneratingJob("out", ["a", "b"], doit)
        ppg.run()

        sl = list(Path("out/by_input").glob("*"))
        sl[0].unlink()
        sl[0].mkdir()
        (sl[0] / "a").write_text("a")
        (sl[0] / "b").write_text("a")
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert isinstance(job.exception, ppg.JobEvaluationFailed)

    def assert_no_rehashing(self):
        org_hash = ppg.hashers.hash_file
        hash_counter = [0]

        def fake_hash(*args):
            hash_counter[0] += 1
            return org_hash(*args)

        ppg.hashers.hash_file = fake_hash
        try:
            ppg.run()
        finally:
            ppg.hashers.hash_file = org_hash
        assert hash_counter[0] == 0

    def test_depends_on_job_gen(self):
        a = ppg.JobGeneratingJob("a", lambda: None)

        def doit(output_files, prefix):
            counter("doit")
            for f in output_files:
                f.write_text(f.name)

        b = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit)
        b.depends_on(a)
        ppg.run()
        assert read("doit") == "1"

        self.assert_no_rehashing()
        assert read("doit") == "1"

    def test_depends_on_various(self):
        class Loader:
            pass

        loader = Loader()

        def doit(output_files, prefix):
            counter("doit")
            for f in output_files:
                f.write_text(f.name)

        def gen():
            job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit)
            Path("input_file").write_text("hello")
            job.depends_on_file("input_file")
            job.depends_on(ppg.DataLoadingJob("shu", lambda: 55))
            job.depends_on(
                ppg.AttributeLoadingJob("shao", loader, "shao", lambda: "shao")
            )
            job.depends_on(ppg.FileGeneratingJob("a", lambda of: of.write_text("a")))

        gen()
        ppg.run()
        self.assert_no_rehashing()

        # the job runs.
        # It just doesn't do the rebild

        # attribute loadin gjob, dataloadingjob, FileInvariant, MultiFileGeneratingJob

    def test_cleanup_multiple(self):
        def doit(output_files, prefix):
            count = str(counter("doit"))
            for f in output_files:
                f.write_text(f.name + count)

        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit, remove_unused=False)
        ppg.run()
        job.depends_on(ppg.ParameterInvariant("b", "b"))
        ppg.run()
        job.depends_on(ppg.ParameterInvariant("c", "c"))
        ppg.run()
        assert len(list(Path("out/done").glob("*"))) == 3
        ppg.new()
        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit, remove_unused=True)
        job.depends_on(ppg.ParameterInvariant("b", "b"))
        job.depends_on(ppg.ParameterInvariant("c", "c"))
        ppg.run()  # job does not run
        assert len(list(Path("out/done").glob("*"))) == 3
        job.depends_on(ppg.ParameterInvariant("d", "d"))
        ppg.run()
        assert len(list(Path("out/done").glob("*"))) == 1

    def test_local_log_usage(self):
        def load():
            fn = (
                ppg.global_pipegraph.history_dir
                / ppg.SharedMultiFileGeneratingJob.log_filename
            )
            return json.loads(fn.read_text())

        def doit(output_files, prefix):
            count = str(counter("doit"))
            for f in output_files:
                f.write_text(f.name + count)

        job = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit, remove_unused=False)
        ppg.run()
        known = load()
        assert (
            Path("out/by_input") / known[str(job.output_dir_prefix)] / "a"
        ).read_text() == "a0"
        job.depends_on_params("shu")
        ppg.run()
        known2 = load()
        assert (
            Path("out/by_input") / known2[str(job.output_dir_prefix)] / "a"
        ).read_text() == "a1"
        assert known != known2

    def test_more_files(self):
        def doit(output_files, prefix):
            count = str(counter("doit"))
            for f in output_files:
                f.write_text(f.name + count)

        a = ppg.SharedMultiFileGeneratingJob("out", ["a"], doit)
        ppg.run()
        assert a["a"].read_text() == "a0"
        assert Path("doit").read_text() == "1"
        ppg.run()
        assert a["a"].read_text() == "a0"
        assert Path("doit").read_text() == "1"
        ppg.new()
        a = ppg.SharedMultiFileGeneratingJob("out", ["a", "b"], doit)
        ppg.run()
        assert Path("doit").read_text() == "2"
        assert a["a"].read_text() == "a1"
        assert a["b"].read_text() == "b1"
        ppg.run()
        assert Path("doit").read_text() == "2"
        assert a["a"].read_text() == "a1"
        assert a["b"].read_text() == "b1"

        ppg.new()
        a = ppg.SharedMultiFileGeneratingJob("out", ["b"], doit)
        ppg.run()
        assert Path("doit").read_text() == "3"
        assert a["b"].read_text() == "b2"

    def test_same_output_dir_multple_jobs(self):
        def doit(output_files, prefix):
            count = str(counter("doit"))
            for f in output_files:
                f.write_text(f.name + count)

        ppg.SharedMultiFileGeneratingJob("out", ["a"], doit)
        with pytest.raises(ppg.JobOutputConflict):
            ppg.SharedMultiFileGeneratingJob("out", ["b"], doit)



from .shared import counter, read
from pathlib import Path
import pypipegraph2 as ppg
import pytest


@pytest.mark.usefixtures("ppg2_per_test")
class TestCallSyntax:
    def test_simple(self):
        a = ppg.FileGeneratingJob("A", lambda of: counter("a") and of.write_text("A"))
        b = ppg.FileGeneratingJob("B", lambda of: counter("b") and of.write_text("B"))
        assert a() == [Path("A")]
        assert read("A") == "A"
        assert not Path("B").exists()
        assert read("a") == "1"
        assert not Path("b").exists()
        b()
        assert read("B") == "B"
        assert read("a") == "1"
        assert read("b") == "1"

    def test_downstream(self):
        a = ppg.MultiFileGeneratingJob(
            {"a": "A"}, lambda of: counter("a") and of["a"].write_text("A")
        )
        b = ppg.FileGeneratingJob("B", lambda of: counter("b") and of.write_text("B"))
        b.depends_on(a)
        assert set(a().keys()) == set(["a"])
        assert read("A") == "A"
        assert not Path("B").exists()
        assert read("a") == "1"
        assert not Path("b").exists()
        b()
        assert read("B") == "B"
        assert read("a") == "1"
        assert read("b") == "1"
        Path("A").unlink()
        b()
        assert read("B") == "B"
        assert read("a") == "2"
        assert read("b") == "1"  # was shielded

    def test_job_is_pruned(self, job_trace_log):
        a = ppg.FileGeneratingJob("A", lambda of: counter("a") and of.write_text("A"))
        b = ppg.FileGeneratingJob("B", lambda of: counter("b") and of.write_text("B"))
        a.prune()
        a()
        assert read("A") == "A"
        assert not Path("B").exists()
        assert read("a") == "1"
        assert not Path("b").exists()
        b()
        assert read("B") == "B"
        assert read("a") == "1"
        assert read("b") == "1"

    def test_plot_job(self):
        import pandas as pd
        import plotnine

        def calc():
            return pd.DataFrame({"X": list(range(0, 100)), "Y": list(range(50, 150))})

        def plot(df):
            p = plotnine.ggplot(df)
            p = p + plotnine.geom_point(plotnine.aes("X", "Y"))
            return p

        of = "A.png"
        p, c, t = ppg.PlotJob(of, calc, plot)
        pout = p()
        assert Path(of).exists()
        assert isinstance(pout, plotnine.ggplot)
        ppg.new()
        p, c, t = ppg.PlotJob(of, calc, plot)
        pout = p()
        assert isinstance(pout, plotnine.ggplot)



#!/usr/bin/env python3
import pypipegraph2 as ppg
from pathlib import Path
import os

run_dir = (Path(__file__).parent / "run_kilojobs").absolute()
run_dir.mkdir(exist_ok=True)
counter = 0


def gen_job(ii):
    return ppg.FileGeneratingJob(f"{ii}.dat", lambda of: of.write_text("h"))


def kilo_jobs():
    os.chdir(run_dir)
    ppg.new()
    final = ppg.FileGeneratingJob("final.dat", lambda of: of.write_text("h"))

    for ii in range(100000):
        final.depends_on(gen_job(ii))
    ppg.run()


if __name__ == "__main__":
    kilo_jobs()


# just doing it: 45s. (183k history)
# noop rerun: 11s
# rerun, rebuild final:  24s


# after dicodon fixes
# noop rerun:   22s
# rerrun with rebuild final:  9s


# 100k
# run : 25m35s. 1.8M (compressed) history. 81MB uncompressed...
# noop run: 1minute.



# demonstrate runtime logging
import pypipegraph2 as ppg
import time
from pathlib import Path
import os
import shutil

p = Path("run/runtimes")
if p.exists():
    shutil.rmtree(p)
Path("run/runtimes").mkdir(exist_ok=True, parents=True)
os.chdir("run/runtimes")


def gen_job(name, runtime):
    def inner(fn):
        time.sleep(runtime)
        fn.write_text(str(runtime))

    return ppg.FileGeneratingJob(name, inner)


gen_job("A1", 1)
gen_job("A.5", 0.5)
gen_job("A2", 2)
gen_job("A5", 5)
ppg.run()
#



from pathlib import Path
import pytest
import pypipegraph2 as ppg
from .shared import write, read


@pytest.mark.usefixtures("ppg2_per_test")
class TestPruning:
    def test_basic_prune(self):
        ppg.FileGeneratingJob("A", lambda of: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda of: write("B", "B"))
        b.prune()
        ppg.run()
        assert Path("A").read_text() == "A"
        assert not Path("B").exists()

    def test_basic_prune2(self):
        a = ppg.FileGeneratingJob("A", lambda of: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda of: write("B", "B"))
        b.depends_on(a)
        b.prune()
        ppg.run()
        assert Path("A").read_text() == "A"
        assert not Path("B").exists()

    def test_basic_prune3(self):
        a = ppg.FileGeneratingJob("A", lambda of: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda of: write("B", "B"))
        c = ppg.FileGeneratingJob("C", lambda of: write("C", "C"))
        d = ppg.FileGeneratingJob("D", lambda of: write("D", "D"))
        b.depends_on(a)
        b.prune()
        c.depends_on(b)  # that is ok, pruning happens after complet build.
        d.depends_on(a)
        ppg.run()
        assert Path("A").read_text() == "A"
        assert Path("D").read_text() == "D"
        assert not Path("B").exists()
        assert not Path("C").exists()
        assert c.prune_reason == b.job_id
        ppg.run()  # just so we recurse_prune again.

    def test_tempfile_not_run_on_prune(self):
        a = ppg.TempFileGeneratingJob("A", lambda of: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda of: write("B", "B" + read("A")))
        b.depends_on(a)
        b.prune()
        ppg.run()
        assert not Path("B").exists()
        assert not Path("A").exists()

    def test_tempfile_still_run_if_needed_for_other(self):
        a = ppg.TempFileGeneratingJob("A", lambda of: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda of: write("B", "B" + read("A")))
        c = ppg.FileGeneratingJob("C", lambda of: write("C", "C" + read("A")))
        b.depends_on(a)
        c.depends_on(a)
        b.prune()
        ppg.run()
        assert not Path("B").exists()
        assert Path("C").exists()
        assert Path("C").read_text() == "CA"
        assert not Path("A").exists()

    def test_basic_prune_unprune(self):
        ppg.FileGeneratingJob("A", lambda of: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda of: write("B", "B"))
        b.prune()
        ppg.run()
        assert Path("A").read_text() == "A"
        assert not Path("B").exists()
        b.unprune()
        ppg.run()
        assert Path("A").read_text() == "A"
        assert read("B") == "B"



import pytest
import threading
from pypipegraph2.parallel import CoreLock


class TestCoreLock:
    def test_single_thread(self):
        mylock = CoreLock(1)
        with pytest.raises(ValueError):
            mylock._acquire(2)
        mylock._acquire(1)
        mylock._release(1)

    def test_multithreading_simple(self):
        mylock = CoreLock(1)
        counter = []
        threads = []

        def inner(c):
            mylock._acquire(1)
            counter.append(c)
            mylock._release(1)

        for i in range(5):
            t = threading.Thread(target=inner, args=(i,))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()
        assert len(counter) == 5
        assert set(counter) == set([0, 1, 2, 3, 4])

    def test_multithreading_two_by_two(self):
        mylock = CoreLock(3)
        counter = []
        threads = []

        def inner(c):
            with mylock.using(2):
                counter.append(c)

        for i in range(5):
            t = threading.Thread(target=inner, args=(i,))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()
        assert len(counter) == 5
        assert set(counter) == set([0, 1, 2, 3, 4])

    def test_multithreading_complex(self):
        mylock = CoreLock(8)
        counter = []
        threads = []

        def inner(c):
            with mylock.using(c % 8 + 1):
                counter.append(c)

        for i in range(8 * 4 + 1):
            t = threading.Thread(target=inner, args=(i,))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()
        assert len(counter) == 33
        assert set(counter) == set(range(33))

    def test_multithreading_dieing(self):
        mylock = CoreLock(1)
        counter = []
        threads = []

        def inner(c):
            with mylock.using(1):
                if c == 1:
                    raise ValueError(
                        "This should show up in the pytest warnings summary"
                    )
                counter.append(c)

        for i in range(5):
            t = threading.Thread(target=inner, args=(i,))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()
        assert len(counter) == 4
        assert set(counter) == set([0, 2, 3, 4])

    def test_aquire_0_raise(self):
        mylock = CoreLock(1)
        with pytest.raises(ValueError):
            mylock._acquire(0)
        with pytest.raises(ValueError):
            with mylock.using(0):
                pass



# -*-tcoding: utf-8 -*-
"""
    Dummy conftest.py for pypipegraph2.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""

# import pytest
import pytest
from pathlib import Path
import shutil
import os
import pypipegraph2 as ppg2
import pypipegraph2.testing.fixtures
import sys
import plotnine  # noqa: F401 so it's available in the plot tests - saves about 10% of runtime
from pypipegraph2.testing.fixtures import job_trace_log

# support code to remove test created files
# only if the test succeedd
# ppg2.util._running_inside_test = True
if "pytest" not in sys.modules:
    raise ValueError("fixtures can only be used together with pytest")


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, "rep_" + rep.when, rep)


ppg2_per_test = ppg2.testing.fixtures.new_pipegraph


@pytest.fixture
def dir_per_test(request):
    """No pipegraph, but seperate directory per test"""
    if request.cls is None:
        target_path = Path(request.fspath).parent / "run" / ("." + request.node.name)
    else:
        target_path = (
            Path(request.fspath).parent
            / "run"
            / (request.cls.__name__ + "." + request.node.name)
        )
    if target_path.exists():  # pragma: no cover
        shutil.rmtree(target_path)
    target_path = target_path.absolute()
    target_path.mkdir()
    old_dir = Path(os.getcwd()).absolute()
    os.chdir(target_path)
    try:

        def np():
            ppg2.util.global_pipegraph = None
            return None

        def finalize():
            if hasattr(request.node, "rep_setup"):

                if request.node.rep_setup.passed and (
                    request.node.rep_call.passed
                    or request.node.rep_call.outcome == "skipped"
                ):
                    try:
                        shutil.rmtree(target_path)
                    except OSError:  # pragma: no cover
                        pass

        request.addfinalizer(finalize)
        ppg2.util.global_pipegraph = None
        yield np()

    finally:
        os.chdir(old_dir)


@pytest.fixture
def create_out_dir(request):
    Path("out").mkdir(exist_ok=True)
    yield


trace_log = job_trace_log


@pytest.fixture
def ppg1_compatibility_test(request):
    import sys
    import pypipegraph as ppg

    ppg2.replace_ppg1()

    if request.cls is None:
        target_path = Path(request.fspath).parent / "run" / ("." + request.node.name)
    else:
        target_path = (
            Path(request.fspath).parent
            / "run"
            / (request.cls.__name__ + "." + request.node.name)
        )
        target_path = target_path.absolute()
    old_dir = Path(os.getcwd()).absolute()
    if old_dir == target_path:
        pass
    else:
        if target_path.exists():  # pragma: no cover
            shutil.rmtree(target_path)

    try:
        first = [False]

        def np(quiet=True, **kwargs):
            if not first[0]:
                Path(target_path).mkdir(parents=True, exist_ok=True)
                os.chdir(target_path)
                Path("logs").mkdir()
                Path("cache").mkdir()
                Path("results").mkdir()
                Path("out").mkdir()
                import logging

                h = logging.getLogger("pypipegraph")
                h.setLevel(logging.WARNING)
                first[0] = True

            if not "resource_coordinator" in kwargs:
                kwargs["resource_coordinator"] = ppg.resource_coordinators.LocalSystem(
                    1, interactive=False
                )
            if not "dump_graph" in kwargs:
                kwargs["dump_graph"] = False
            if not "quiet" in kwargs:
                kwargs["quiet"] = quiet
            ppg.new_pipegraph(**kwargs)
            ppg.util.global_pipegraph.result_dir = Path("results")
            g = ppg.util.global_pipegraph
            g.new_pipegraph = np
            return g

        def finalize():
            if hasattr(request.node, "rep_setup"):

                if request.node.rep_setup.passed and (
                    request.node.rep_call.passed
                    or request.node.rep_call.outcome == "skipped"
                ):
                    try:
                        if not hasattr(ppg.util.global_pipegraph, "test_keep_output"):
                            if "--profile" not in sys.argv:
                                shutil.rmtree(target_path)
                    except OSError:  # pragma: no cover
                        pass

        request.addfinalizer(finalize)
        yield np()

    finally:
        os.chdir(old_dir)
        ppg2.unreplace_ppg1()



import os
import time
import subprocess
import sys
import pytest
import pypipegraph2 as ppg
from pathlib import Path
from .shared import write, read, Dummy, counter


@pytest.mark.usefixtures("ppg2_per_test", "create_out_dir")
class TestResourceCoordinator:
    def test_jobs_that_need_all_cores_are_spawned_one_by_one(self):
        # we'll determine this by the start respective end times..
        ppg.new(
            cores=3,
        )

        def a(of):
            write(of, "A")
            time.sleep(0.1)

        def b(of):
            write(of, "B")
            time.sleep(0.1)

        jobA = ppg.FileGeneratingJob("out/A", a, resources=ppg.Resources.AllCores)
        jobB = ppg.FileGeneratingJob("out/B", b, resources=ppg.Resources.AllCores)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        if jobA.start_time < jobB.start_time:
            first_job = jobA
            second_job = jobB
        else:
            first_job = jobB
            second_job = jobA
        min_start = min(jobA.start_time, jobB.start_time)
        print(
            "times",
            first_job.start_time - min_start,
            first_job.stop_time - min_start,
            second_job.start_time - min_start,
            second_job.stop_time - min_start,
        )
        if jobA.start_time is None:
            raise ValueError("JobA did not run")
        assert first_job.stop_time < second_job.start_time

    def test_jobs_concurrent_jobs_run_concurrently(self):
        # we'll determine this by the start respective end times..
        import time

        ppg.new(
            cores=2,
        )
        jobA = ppg.FileGeneratingJob(
            "out/A", lambda of: write(of, "A"), resources=ppg.Resources.AllCores
        )
        jobB = ppg.FileGeneratingJob(
            "out/B", lambda of: write(of, "B"), resources=ppg.Resources.AllCores
        )
        now = time.time()
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        if jobA.start_time < jobB.start_time:
            first_job = jobA
            second_job = jobB
        else:
            first_job = jobB
            second_job = jobA
        print(
            "times",
            first_job.start_time - now,
            first_job.stop_time - now,
            second_job.start_time - now,
            second_job.stop_time - now,
        )
        if jobA.start_time is None:
            raise ValueError("JobA did not run")
        assert first_job.stop_time - now > second_job.start_time - now

    def test_multiple_all_cores_blocking_single_jobs(self):
        pass  # todo


class CantDepickle:
    """A class that can't be depickled (throws a type error,
    just like the numpy.maskedarray does occacionally)"""

    def __getstate__(self):
        return {"shu": "5"}

    def __setstate__(self, state):
        print(state)
        raise TypeError("I can be pickled, but not unpickled")


@pytest.mark.usefixtures("ppg2_per_test")
class TestingTheUnexpectedTests:
    def test_job_exiting_python(self):
        def dies(of):
            import sys
            from loguru import logger

            logger.info("Now terminating child python")
            sys.exit(5)

        ppg.FileGeneratingJob("out/A", dies)
        with pytest.raises(ppg.JobsFailed):
            # ppg.util.global_pipegraph.rc.timeout = 1
            ppg.run()
        assert not (Path("out/A").exists())
        error = ppg.global_pipegraph.last_run_result["out/A"].error
        assert isinstance(error.args[0], ppg.JobDied)
        assert error.args[0].args[2] == 5

    def test_job_exiting_python_stdout_stderr_logged(self):
        def dies(of):
            import sys

            # logging.info("Now terminating child python")
            print("hello")
            sys.stderr.write("I am stderr\n")
            sys.stdout.flush()
            sys.exit(5)

        fg = ppg.FileGeneratingJob("out/A", dies)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
            ppg.run()
        assert not (Path("out/A").exists())
        error = ppg.global_pipegraph.last_run_result["out/A"].error
        assert isinstance(error.args[0], ppg.JobDied)
        assert error.args[0].args[2] == 5
        assert fg.stdout == "hello\n"
        assert fg.stderr == "I am stderr\n"

    def test_job_getting_killed_python_stdout_stderr_logged(self):
        def dies(of):
            import sys
            import signal

            # logging.info("Now terminating child python")
            print("hello")
            sys.stderr.write("I am stderr\n")
            sys.stdout.flush()
            sys.stderr.flush()
            os.kill(os.getpid(), signal.SIGKILL)

        fg = ppg.FileGeneratingJob("out/A", dies)
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
            ppg.run()
        assert not (Path("out/A").exists())
        error = ppg.global_pipegraph.last_run_result["out/A"].error
        assert isinstance(error.args[0], ppg.JobDied)
        assert error.args[0].args[2] == -9
        assert fg.stdout == "hello\n"
        assert fg.stderr == "I am stderr\n"

    def testing_import_does_not_hang(self):  # see python issue22853
        old_dir = os.getcwd()
        os.chdir(Path(__file__).parent)
        p = subprocess.Popen(
            [sys.executable, "_import_does_not_hang.py"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        print(stdout, stderr)
        assert b"OK" in stdout
        os.chdir(old_dir)

    def test_older_jobs_added_back_to_new_pipegraph(self, create_out_dir):
        a = ppg.FileGeneratingJob("out/A", lambda of: write(of, "a"))
        # ppg.run()
        ppg.new()
        b = ppg.FileGeneratingJob("out/B", lambda of: write(of, "b"))
        with pytest.raises(KeyError):
            a.depends_on(b)
        with pytest.raises(KeyError):
            b.depends_on(a)
        a.readd()
        a.depends_on(b)
        assert "out/B" in ppg.global_pipegraph.jobs
        assert "out/A" in ppg.global_pipegraph.jobs
        # with pytest.raises(ppg.PyPipeGraphError):


@pytest.mark.usefixtures("ppg2_per_test")
class TestPathLib:
    def test_multifilegenerating_job_requires_string_filenames(self):
        import pathlib

        x = lambda of: 5  # noqa:E731
        ppg.MultiFileGeneratingJob(["a"], x)
        ppg.MultiFileGeneratingJob([pathlib.Path("a")], x)

        with pytest.raises(TypeError):
            ppg.MultiFileGeneratingJob([0])

        with pytest.raises(TypeError):
            ppg.MultiFileGeneratingJob([b"a"])  # bytes is not a string type

    def test_accepts(self):
        import pathlib

        write("aaa", "hello")
        write("bbb", "hello")
        write("ccc", "hello")
        a = ppg.FileInvariant(pathlib.Path("aaa"))
        b = ppg.FileGeneratingJob(
            pathlib.Path("b"),
            lambda of: write(of, "bb" + read("aaa") + read("bbb") + read("ccc")),
        )
        b.depends_on(a)

        dd = Dummy()

        def mf(ofs):
            ofs[0].write_text("cc" + read("g"))
            ofs[1].write_text("dd" + read("h") + dd.attr)
            ofs[2].write_text("ee" + read("i") + read("j") + read("k"))

        c = ppg.MultiFileGeneratingJob([pathlib.Path("c"), "d", pathlib.Path("e")], mf)
        c.depends_on(b)
        d = ppg.FunctionInvariant(pathlib.Path("f"), lambda x: x + 1)
        c.depends_on(d)
        e = ppg.ParameterInvariant(pathlib.Path("c"), "hello")
        c.depends_on(e)
        f = ppg.TempFileGeneratingJob(pathlib.Path("g"), lambda of: write(of, "gg"))
        c.depends_on(f)

        def tmf(ofs):
            ofs[0].write_text("hh")
            ofs[1].write_text("ii")

        g = ppg.MultiTempFileGeneratingJob([pathlib.Path("h"), "i"], tmf)
        c.depends_on(g)

        def tpf(ofs):
            assert len(ofs) == 2
            write(ofs[0], "jjjj")
            write(ofs[1], "kkkk")

        with pytest.raises(ValueError):
            h = ppg.MultiTempFileGeneratingJob(
                [pathlib.Path("j"), "k", pathlib.Path("k")], tpf
            )
        h = ppg.MultiTempFileGeneratingJob([pathlib.Path("j"), "k"], tpf)
        c.depends_on(h)

        i = ppg.CachedDataLoadingJob(
            pathlib.Path("l"), lambda: write("l", "llll"), lambda res: res
        )
        with pytest.raises(TypeError):
            c.depends_on(i)
        c.depends_on(i.load)

        m = ppg.CachedAttributeLoadingJob(pathlib.Path("m"), dd, "attr", lambda: "55")
        c.depends_on(m.load)
        ppg.run()
        assert read("aaa") == "hello"
        assert read("b") == "bbhellohellohello"
        assert read("c") == "ccgg"
        assert read("d") == "ddhh55"
        assert read("e") == "eeiijjjjkkkk"
        assert not (Path("g").exists())
        assert not (Path("h").exists())
        assert not (Path("i").exists())
        assert not (Path("j").exists())


def test_fixture_without_class(ppg2_per_test):
    # just to make sure the ppg2_per_test fixture does what it's supposed to if you're not using a class
    import pathlib

    assert "run/.test_fixture_without_class" in str(pathlib.Path(".").absolute())


def test_job_or_filename(ppg2_per_test):
    a, dep_a = ppg.util.job_or_filename("out/A")
    assert a == Path("out/A")
    assert len(dep_a) == 1
    assert isinstance(dep_a[0], ppg.FileInvariant)
    j = ppg.FileGeneratingJob("out/B", lambda of: None)
    b, dep_b = ppg.util.job_or_filename(j)
    assert b == Path("out/B")
    assert dep_b[0] is j
    assert len(dep_b) == 1

    c, dep_c = ppg.util.job_or_filename(None)
    assert c is None
    assert not dep_c


@pytest.mark.xfail()  # todo
def test_interactive_import(ppg2_per_test):
    # just so at least the import part of interactive is under coverage
    import pypipegraph2.interactive  # noqa:F401


def test_dataloading_job_changing_cwd(ppg2_per_test):
    from pathlib import Path

    os.mkdir("shu")

    def load():
        os.chdir("shu")
        Path("b").write_text("world")
        return 55

    a = ppg.FileGeneratingJob("a", lambda of: of.write_text("hello"))
    b = ppg.DataLoadingJob("b", load)
    a.depends_on(b)
    with pytest.raises(ppg.JobsFailed):
        ppg.run()
    assert isinstance(
        ppg.global_pipegraph.last_run_result["b"].error.args[0], ppg.JobContractError
    )


def test_job_generating_job_changing_cwd(ppg2_per_test):
    from pathlib import Path

    os.mkdir("shu")

    def load():
        os.chdir("shu")
        Path("b").write_text("world")
        return 55

    a = ppg.FileGeneratingJob("a", lambda of: Path("a").write_text("hello"))
    b = ppg.JobGeneratingJob("b", load)
    a.depends_on(b)
    with pytest.raises(ppg.JobsFailed):
        ppg.run()
    assert isinstance(
        ppg.global_pipegraph.last_run_result["b"].error.args[0], ppg.JobContractError
    )


def test_capturing_locals_when_they_have_throwing_str(ppg2_per_test):
    class NoStr:
        def __str__(self):
            raise ValueError("Cant string this")

    def inner(of):
        a = NoStr()
        raise ValueError("expected")  # trace check

    j = ppg.FileGeneratingJob("a", inner)
    with pytest.raises(ppg.JobsFailed):
        ppg.run()
    assert "expected" in str(j.exception)
    assert "trace check" in str(j.stack_trace)  # we captured teh relevant line


def test_cache_dir(ppg2_per_test):
    ppg.new(cache_dir="shu")
    assert Path("shu").exists()
    ppg.new(cache_dir=None)
    a = ppg.FileGeneratingJob("a", lambda of: of.write_text("A"))
    ppg.run()
    assert Path("a").read_text() == "A"


class TestCleanup:
    def test_error_cleanup(self, ppg2_per_test):
        import pypipegraph2 as ppg2

        ppg2_per_test.new(log_retention=2)
        assert ppg2.global_pipegraph.log_retention == 2

        def fail(of):
            raise ValueError()

        job = ppg.FileGeneratingJob("A", fail)
        ec = []  # counters
        lc = []
        with pytest.raises(ppg.JobsFailed):
            ppg.run(print_failures=False)
        ec.append(len(list(ppg.global_pipegraph.error_dir.glob("*"))))
        lc.append(len(list(ppg.global_pipegraph.log_dir.glob("*.log"))))
        with pytest.raises(ppg.JobsFailed):
            ppg.run(print_failures=False)
        ec.append(len(list(ppg.global_pipegraph.error_dir.glob("*"))))
        lc.append(len(list(ppg.global_pipegraph.log_dir.glob("*.log"))))
        with pytest.raises(ppg.JobsFailed):
            ppg.run(print_failures=False)
        # we keep  log_retention old ones + the current one
        ec.append(len(list(ppg.global_pipegraph.error_dir.glob("*"))))
        lc.append(len(list(ppg.global_pipegraph.log_dir.glob("*.log"))))

        with pytest.raises(ppg.JobsFailed):
            ppg.run(print_failures=False)
        assert ppg.global_pipegraph.log_file.exists()
        ec.append(len(list(ppg.global_pipegraph.error_dir.glob("*"))))
        lc.append(len(list(ppg.global_pipegraph.log_dir.glob("*.log"))))
        assert ec == [
            1 + 1,
            2 + 1,
            3 + 1,  # latest
            3 + 1,
        ]
        assert lc == [
            1,
            2,
            3,  # latest exluded by *.log
            3,
        ]


def test_exploding_core_lock_captured(ppg2_per_test):
    """We had an issue where an exception in acquiring the core lock led to the thread dying
    and the graph runner hanging
    """
    a = ppg.FileGeneratingJob(
        "a", lambda of: of.write_text("a"), resources=ppg.Resources._RaiseInCoreLock
    )
    with pytest.raises(ppg.FatalGraphException):
        ppg.run()
    assert "Count == 0" in str(a.exception)


def test_exploding_resources_to_number(ppg2_per_test):
    """We had an issue where an exception in acquiring the core lock led to the thread dying
    and the graph runner hanging
    """
    a = ppg.FileGeneratingJob(
        "a", lambda of: of.write_text("a"), resources=ppg.Resources._RaiseInToNumber
    )
    with pytest.raises(ppg.FatalGraphException):
        ppg.run()
    assert "Not a Resource" in str(a.exception)


@pytest.mark.usefixtures("ppg2_per_test")
class TestModifyDag:
    def test_2_fg_one_dl(self):
        from pypipegraph2.util import log_info

        parts = []
        a1 = ppg.FileGeneratingJob(
            "a1", lambda of: of.write_text(str(parts)), depend_on_function=False
        )
        a2 = ppg.FileGeneratingJob(
            "a2",
            lambda of: of.write_text(Path("a1").read_text() + str(parts)),
            depend_on_function=False,
        )
        a2.depends_on(a1)
        b = ppg.DataLoadingJob(
            "b", lambda *args: parts.append(1), depend_on_function=False
        )
        a1.depends_on(b)
        a2.depends_on(b)
        log_info("now run")
        ppg.run()
        assert read("a1") == "[1]"
        assert read("a2") == "[1][1]"

    def test_2_fg_one_dl_failing(pself, ppg2_per_test, job_trace_log):
        from pypipegraph2.util import log_info

        parts = []
        a1 = ppg.FileGeneratingJob(
            "a1", lambda of: of.write_text("a1"), depend_on_function=False
        )
        a2 = ppg.FileGeneratingJob(
            "a2",
            lambda of: of.write_text(Path("a2").read_text() + str(parts)),
            depend_on_function=False,
        )
        a2.depends_on(a1)
        b = ppg.DataLoadingJob(
            "b", lambda: _no_such_thing, depend_on_function=False  # noqa: F821
        )
        a1.depends_on(b)
        a2.depends_on(b)
        log_info("now run")
        with pytest.raises(ppg.JobsFailed):
            ppg.run()
        assert type(b.exception) is NameError
        assert "Upstream" in str(a1.exception)
        assert "Upstream" in str(a2.exception)
        assert not Path("a1").exists()
        assert not Path("a2").exists()

    def test_2_fg_one_temp(ppg2_per_test):
        from pypipegraph2.util import log_info

        parts = []
        a1 = ppg.FileGeneratingJob(
            "a1",
            lambda of: of.write_text(Path("b").read_text() + "a1"),
            depend_on_function=False,
        )
        a2 = ppg.FileGeneratingJob(
            "a2",
            lambda of: of.write_text(
                Path("b").read_text() + Path("a1").read_text() + "a2"
            ),
            depend_on_function=False,
        )
        a2.depends_on(a1)
        b = ppg.TempFileGeneratingJob(
            "b", lambda of: of.write_text("b"), depend_on_function=False
        )
        a1.depends_on(b)
        a2.depends_on(b)

        log_info("now run")
        ppg.run()
        assert read("a1") == "ba1"
        assert read("a2") == "bba1a2"
        assert not Path("b").exists()


def test_prevent_absolute_paths(ppg2_per_test):
    ppg2_per_test.new(prevent_absolute_paths=True)
    with pytest.raises(ValueError):
        ppg.FileGeneratingJob("/tmp/absolute", lambda of: of.write_text("a"))
    ppg2_per_test.new(prevent_absolute_paths=False)
    ppg.FileGeneratingJob("/tmp/absolute", lambda of: of.write_text("a"))
    ppg2_per_test.new(prevent_absolute_paths=True)


def test_broken_case_from_delayeddataframe(ppg2_per_test):
    out = {}

    def store(key, value):
        out[key] = value

    a = ppg.CachedDataLoadingJob(
        "a", lambda: "a", lambda value: counter("A") and store("a", value)
    )
    event = ppg.CachedDataLoadingJob(
        "event",
        lambda: counter("EVENT") and "event",
        lambda value: store("event", value),
    )
    event2 = ppg.CachedDataLoadingJob(
        "event2",
        lambda: counter("EVENT2") and "event2",
        lambda value: store("event2", value),
    )
    anno_sequence = ppg.CachedDataLoadingJob(
        "anno_sequence",
        lambda: counter("ANNO_SEQUENCE") and "anno_sequence",
        lambda value: store("anno_sequence", value),
    )
    event_seq = ppg.DataLoadingJob(
        "event_seq", lambda: store("event_seq", out["event"] + out["anno_sequence"])
    )
    event2_seq = ppg.DataLoadingJob(
        "event2_seq", lambda: store("event2_seq", out["event2"] + out["event_seq"])
    )
    force_load = ppg.JobGeneratingJob("force_load", lambda: None)

    anno_sequence.calc.depends_on(a.load)
    anno_sequence.load.depends_on(a.load)
    event.calc.depends_on(a.load)

    event_seq.depends_on(event.load)
    event_seq.depends_on(anno_sequence.load)

    event2.calc.depends_on(event.load)

    event2_seq.depends_on(event_seq, event2.load)

    force_load.depends_on(event.load, event2.load, event2_seq, event_seq)
    ppg.run()
    assert out["event2_seq"] == "event2" + "event" + "anno_sequence"
    assert read("EVENT") == "1"
    assert read("EVENT2") == "1"
    assert read("ANNO_SEQUENCE") == "1"
    ppg.run()
    assert read("EVENT") == "1"
    assert read("EVENT2") == "1"
    assert read("ANNO_SEQUENCE") == "1"


def test_strict_mode_two_jobs_same_id(ppg2_per_test):
    ppg.new(run_mode=ppg.RunMode.CONSOLE)
    assert ppg.global_pipegraph.run_mode.is_strict()
    a = lambda of: None  # noqa: E731
    ppg.FileGeneratingJob("a", a, depend_on_function=False)
    ppg.FileGeneratingJob("a", a, depend_on_function=False)
    # the testing is actually on the FunctionInvariant :(
    ppg.FileGeneratingJob("a", a, depend_on_function=True)
    with pytest.raises(ppg.JobRedefinitionError):
        ppg.FileGeneratingJob(
            "a", lambda of: of.write_text("h"), depend_on_function=True
        )


def test_calling_function_difference():
    def gen(x):
        return lambda: x + 5

    a = gen(5)
    b = gen(10)
    r = ppg.FunctionInvariant.debug_function_differences(a, b)
    assert "The function closures differed." in r
    assert (
        ppg.FunctionInvariant.debug_function_differences(None, None) == "No difference"
    )
    assert (
        ppg.FunctionInvariant.debug_function_differences(gen, gen)
        == "The functions were identical"
    )


def test_spawned_processes_get_killed_on_abort(ppg2_per_test, job_trace_log):
    import psutil

    ppg.new(cores=2)

    # this process doesn't get reaped
    p = subprocess.Popen("cat", stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    assert p.pid

    def long_running_proc():
        p = subprocess.Popen(["sleep", "5"], stdin=subprocess.PIPE)
        Path("pid").write_text(str(p.pid))
        ppg.global_pipegraph.runner.abort()
        p.communicate()
        Path("done").write_text("done")

    a = ppg.DataLoadingJob("a", long_running_proc)
    b = ppg.FileGeneratingJob("b", lambda of: of.write_text(str(of))).depends_on(a)
    with pytest.raises(ppg.JobsFailed):
        ppg.run()
    assert not Path("done").exists()
    time.sleep(1)
    assert psutil.pid_exists(p.pid)
    pid = int(Path("pid").read_text())
    ok = False
    try:
        ok = psutil.Process(pid).status() == "zombie"
    except psutil.NoSuchProcess:
        ok = True
    if not ok:
        # subprocess.check_call(["ps", "xf", "-o", "pid,ppid,pgid,cmd"])
        raise ValueError("Process was neither zombie (since parent is dead), nor gone.")

    p.communicate()


def test_spawned_processes_get_killed_on_catastrophic_process_failure(
    ppg2_per_test, job_trace_log
):
    import psutil

    path = Path(__file__).parent / "test_parent_termination_kills_children"
    p = subprocess.Popen(["python", path / "stage2.py"])
    p.communicate()

    for proc in psutil.process_iter(["cmdline"]):
        if "ppg2_test_parallel_sentinel" in proc.info["cmdline"]:
            raise ValueError(
                "children not killed, found a ppg2_test_parallel_sentinel",
                "pid",
                proc.pid,
                "ppid",
                proc.ppid(),
            )
        if "sleep" in proc.info["cmdline"] and "5.51234" in proc.info["cmdline"]:
            raise ValueError(
                "children not killed, found a sleep",
                "pid",
                proc.pid,
                "ppid",
                proc.ppid(),
            )


def test_tempmultifilegen_changing_output_but_had_been_validated(ppg2_per_test):
    import random

    def bad(files):
        files[0].write_text("hello")
        files[1].write_text("world" + str(random.random()))

    tmfg = ppg.MultiTempFileGeneratingJob(["a", "b"], bad)
    c = ppg.FileGeneratingJob("c", lambda of: of.write_text(Path("a").read_text()))
    c.depends_on("a")
    ppg.run()
    ppg.new()
    tmfg = ppg.MultiTempFileGeneratingJob(["a", "b"], bad)
    c = ppg.FileGeneratingJob("c", lambda of: of.write_text(Path("a").read_text()))
    c.depends_on("a")
    d = ppg.FileGeneratingJob("d", lambda of: of.write_text(Path("a").read_text()))
    d.depends_on("a")
    with pytest.raises(ppg.JobsFailed):
        ppg.run()
    error = ppg.global_pipegraph.last_run_result[tmfg.job_id].error
    assert "changed output" in error and "ephemeral" in error.lower()



class PPGException(Exception):
    """Baseclass for all our exceptions"""

    pass


class FatalGraphException(PPGException):
    """Could not finish executing the graph completly"""

    pass


class NotADag(FatalGraphException):
    """Your graph is not acyclic"""

    pass


class JobOutputConflict(ValueError):
    """Multiple jobs with overlapping (but not identical) outputs were defined"""

    pass


class JobContractError(PPGException):
    """A job did not do what it was supposed to do"""

    pass


class JobDied(PPGException):
    pass


class JobRedefinitionError(ValueError):
    pass


class JobEvaluationFailed(PPGException):
    """For some reason, we could not decide on whether to execute this job"""

    pass


class RunFailed(FatalGraphException):
    """The execution failed outside of the scope of a single job"""

    pass


class JobsFailed(RunFailed):
    """one or more jobs failed"""

    def __init__(self, msg, exceptions):
        super().__init__(msg)
        self.exceptions = exceptions


class RunFailedInternally(RunFailed):
    """There is a bug in pypipegraph2"""

    def __init__(self, *args, **kwargs):
        super().__init__(
            "RunFailedInternally: Due to some bug in the graph-running, we could not finish running. File a bug report.",
            *args,
            **kwargs,
        )


class _RunAgain(PPGException):
    """Internal to signal rerun"""

    pass


class JobError(PPGException):
    """Wrapper around the exceptions thrown by random jobs"""

    def __str__(self):
        return (
            ("ppg.JobError:\n")
            + (f"{self.args[0].__class__.__name__}: {self.args[0]}\n")
            # + (f"\tTraceback: {self.args[1]}\n")
            + ("")
        )

    def __repr__(self):
        return str(self)


class JobCanceled(PPGException):
    pass


class HistoryLoadingFailed(FatalGraphException):
    pass



# check if 'stalling' occurs - ie.
# will having to many all_cores jobs ready to run
# prevent all single cores to run from in parallel
import pypipegraph2 as ppg
import time
from pathlib import Path
import logging
import os
import shutil

p = Path("run/stall")
if p.exists():
    shutil.rmtree(p)
p.mkdir(exist_ok=True, parents=True)
os.chdir(p)

timeout = 2
jobcount = 5
ppg.new(log_level=logging.INFO, cores=5)


def all_cores(ii):
    def inner(of):
        of.write_text(str(time.time()))
        time.sleep(timeout)

    return ppg.FileGeneratingJob(
        f"all_cores{ii}",
        inner,
        resources=ppg.Resources.AllCores,
        depend_on_function=False,
    )


def single_cores(ii):
    def inner(of):
        of.write_text(str(time.time()))
        time.sleep(timeout)

    return ppg.FileGeneratingJob(
        f"single{ii}",
        inner,
        resources=ppg.Resources.SingleCore,
        depend_on_function=False,
    )


sc = [single_cores(ii) for ii in range(jobcount)]
ac = [all_cores(ii) for ii in range(jobcount)]

start = time.time()
ppg.run()
stop = time.time()

print(
    f"took {stop-start:.2f}. If this is close to {timeout * jobcount}, all is well. "
    "If it's more, we have an ordering problem"
)



import pypipegraph2 as ppg

should_len = 1024 * 1024


def inner(*args):
    # df = pd.DataFrame()
    # df['shu']
    raise ValueError("x " * (should_len // 2))


ppg.FileGeneratingJob("b", lambda of: 5).depends_on(ppg.DataLoadingJob("A", inner))
ppg.run()
#



"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from pathlib import Path
import pytest
import pypipegraph as ppg
from .shared import write, read


@pytest.mark.usefixtures("ppg1_compatibility_test")
class TestPruning:
    def test_basic_prune(self):
        ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda: write("B", "B"))
        b.prune()
        ppg.run_pipegraph()
        assert Path("A").read_text() == "A"
        assert not Path("B").exists()

    def test_basic_prune2(self):
        a = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda: write("B", "B"))
        b.depends_on(a)
        b.prune()
        ppg.run_pipegraph()
        assert Path("A").read_text() == "A"
        assert not Path("B").exists()

    def test_basic_prune3(self):
        a = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda: write("B", "B"))
        c = ppg.FileGeneratingJob("C", lambda: write("C", "C"))
        d = ppg.FileGeneratingJob("D", lambda: write("D", "D"))
        b.depends_on(a)
        b.prune()
        c.depends_on(b)  # that is ok, pruning happens after complet build.
        d.depends_on(a)
        ppg.run_pipegraph()
        assert Path("A").read_text() == "A"
        assert Path("D").read_text() == "D"
        assert not Path("B").exists()
        assert not Path("C").exists()
        # assert c._pruned == b.job_id ppg3
        assert c.prune_reason == b.job_id

    @pytest.mark.skip  # no FinalJob in ppg2
    def test_pruning_does_not_prune_final_jobs(self):
        ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda: write("B", "B"))
        ppg.FinalJob("shu", lambda: write("C", "C"))
        b.prune()
        ppg.run_pipegraph()
        assert Path("A").read_text() == "A"
        assert Path("C").read_text() == "C"
        assert not Path("B").exists()

    @pytest.mark.skip  # no FinalJob in ppg2
    def test_pruning_final_jobs_directly(self):
        ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        ppg.FileGeneratingJob("B", lambda: write("B", "B"))
        c = ppg.FinalJob("shu", lambda: write("C", "C"))
        c.prune()
        ppg.run_pipegraph()
        assert Path("A").read_text() == "A"
        assert Path("B").read_text() == "B"
        assert not Path("C").exists()

    def test_tempfile_not_run_on_prune(self):
        a = ppg.TempFileGeneratingJob("A", lambda: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda: write("B", "B" + read("A")))
        b.depends_on(a)
        b.prune()
        ppg.run_pipegraph()
        assert not Path("B").exists()
        assert not Path("A").exists()

    def test_tempfile_still_run_if_needed_for_other(self):
        a = ppg.TempFileGeneratingJob("A", lambda: write("A", "A"))
        b = ppg.FileGeneratingJob("B", lambda: write("B", "B" + read("A")))
        c = ppg.FileGeneratingJob("C", lambda: write("C", "C" + read("A")))
        b.depends_on(a)
        c.depends_on(a)
        b.prune()
        ppg.run_pipegraph()
        assert not Path("B").exists()
        assert Path("C").exists()
        assert Path("C").read_text() == "CA"
        assert not Path("A").exists()



import pypipegraph2 as ppg
import pytest


@pytest.mark.usefixtures("ppg2_per_test")
class TestBuildInCompabilty:
    def test_invariant_build_in_function(self):
        a = ppg.FunctionInvariant("test", sorted).run(None, None)["FItest"]["source"]
        assert a == "<built-in function sorted>"


_cytohn_func_counter = 0


@pytest.mark.usefixtures("ppg2_per_test")
class TestCythonCompability:
    def source_via_func_invariant(self, name, func):
        global _cytohn_func_counter
        _cytohn_func_counter = +1
        r = ppg.FunctionInvariant(name + str(_cytohn_func_counter), func).run(
            None, None
        )
        print(r)
        return r["FI" + name + str(_cytohn_func_counter)]["source"]

    def test_just_a_function(self):
        import cython

        src = """
def a():
    '''single line docstring'''
    return 1

def b():
    '''Multi
    line
    docstring
    '''


    shu = 55

    return shu
"""
        func = cython.inline(src)["a"]
        func2 = cython.inline(src)["b"]
        actual = self.source_via_func_invariant("a", func)
        should = """    def a():
        return 1"""
        assert actual == should

        actual = self.source_via_func_invariant("b", func2)
        should = '    def b():\n        shu = 55\n    \n        return shu'
        assert actual == should

        ppg.FunctionInvariant("a", func)  # not a redefinition
        with pytest.raises(ppg.JobRedefinitionError):
            ppg.FunctionInvariant("a", func2)  # cython vs cython
        with pytest.raises(ppg.JobRedefinitionError):
            ppg.FunctionInvariant("a", lambda: 1)  # cython vs python
        ppg.FunctionInvariant("b", lambda: 45)
        with pytest.raises(ppg.JobRedefinitionError):
            ppg.FunctionInvariant("b", func2)  # python vs cython

    def test_just_a_function_with_docstring(self):
        import cython

        src = (
            """
def a():
    ''' a() is used
    to do nothing
    '''
    return 1

"""
            '''def b():
    """ b() is used
    to do nothing as well
    """
    return 5
'''
        )
        func = cython.inline(src)["a"]
        actual = self.source_via_func_invariant("a", func)
        should = """    def a():
        return 1"""
        assert actual == should

    def test_nested_function(self):
        import cython

        src = """
def a():
    def b():
        return 1
    return b

def c():
    return 5
"""
        func = cython.inline(src)["a"]()
        actual = self.source_via_func_invariant("a", func)
        should = """        def b():
            return 1"""
        assert actual == should

    def test_class(self):
        import cython

        src = """
class A():
    def b(self):
        return 55

def c():
    return 5
"""

        func = cython.inline(src)["A"]().b
        actual = self.source_via_func_invariant("a", func)
        should = """        def b(self):
            return 55"""
        assert actual == should

    def test_class_inner_function(self):
        import cython

        src = """
class A():
    def b(self):
        def c():
            return 55
        return c

def d():
    return 5
"""

        func = cython.inline(src)["A"]().b()
        actual = self.source_via_func_invariant("a", func)
        should = """            def c():
                return 55"""
        assert actual == should



from pathlib import Path
from xxhash import xxh3_128


def hash_file(path: Path):
    """delegate to a fast and somewhat collision resistant hash function"""
    # I profiled a bunch of hash functions
    # and xx3 and spooky were the fastest 128bit hashers
    # (we want 128 bit to prevent collisions).
    # single core, spooky seemed a bit faster
    # but the xxhash implementation releases the gil
    # when passed more than 100kb (otherwise it's a
    # faster *not* to acquire the lock!)
    hasher = xxh3_128()
    # if known_st_size is None:
    # known_st_size = path.stat().st_size
    # we are not acquiring the core lock here.
    # why? because this is essentially always
    # limited by the read-bandwidth, not the
    # cpu.
    # (even on a very fast Samsung EVO equivalent SSD
    # (about 3gb/s), doing it from memory is 4 times faster)
    # so we shouldn't be stalling everything else much
    # (except for memory bandwidth. oh well, at least
    # it should not go into swap with the tiny buffer we use here)
    with open(path, "rb") as op:
        block = op.read(1024 * 512)
        while block:
            hasher.update(block)
            block = op.read(1024 * 512)
    stat = path.stat()

    return {
        "hash": hasher.hexdigest(),
        "mtime": int(stat.st_mtime),
        "size": stat.st_size,
    }


def hash_bytes(input: bytes):
    hasher = xxh3_128()
    hasher.update(input)
    return hasher.hexdigest()


def hash_str(input: str):
    return hash_bytes(input.encode("utf-8"))



from pathlib import Path
import types
import inspect
import os
import sys
import logging
import pypipegraph as ppg1
import pypipegraph.testing
import pypipegraph.testing.fixtures
import pypipegraph2 as ppg2
import pypipegraph2.testing
import pypipegraph2.testing.fixtures
import wrapt
import importlib
from .util import (
    # log_info, log_error, log_warning, log_debug,
    log_job_trace,
)


old_entries = {}
old_modules = {}
patched = False

exception_map = {
    "RuntimeError": "JobsFailed",
    "JobContractError": "JobContractError",
    "PyPipeGraphError": "FatalGraphException",
    "CycleError": "NotADag",
    "JobDiedException": "JobDied",
    "RuntimeException": "RunFailedInternally",
}


def replace_ppg1():
    """Turn all ppg1 references into actual ppg2
    objects.
    Best effort, but the commonly used API should be well supported.
    Try to do this before anything imports ppg1.
    """

    global patched
    if patched:
        return
    for x in dir(ppg1):
        old_entries[x] = getattr(ppg1, x)
        delattr(ppg1, x)
    for module_name, replacement in {
        "pypipegraph.job": job,
        "pypipegraph.testing": ppg2.testing,
        "pypipegraph.testing.fixtures": ppg2.testing.fixtures,
    }.items():
        if not module_name in sys.modules:
            importlib.import_module(module_name)

        old_modules[module_name] = sys.modules[module_name]
        sys.modules[module_name] = replacement

    # ppg1.__name__ == "pypipegraph2"
    # ppg1.__file__ == __file__
    # ppg1.__path__ == __path__
    # ppg1.__loader__ == __loader__
    # ppg1.__spec__ == __spec__
    # ppg1.__version__ == ppg2.__version__
    ppg1.__doc__ == """ppg1->2 compatibility layer.
Supports the commonly used the old ppg1 API
with ppg2 objects. Aspires to be a drop-in replacement.
    """

    for old, new in exception_map.items():
        setattr(ppg1, old, getattr(ppg2.exceptions, new))

    # invariants

    ppg1.ParameterInvariant = ParameterInvariant
    ppg1.FileInvariant = FileInvariant
    ppg1.FileTimeInvariant = FileInvariant
    ppg1.RobustFileChecksumInvariant = FileInvariant
    ppg1.FileChecksumInvariant = FileInvariant
    ppg1.FunctionInvariant = FunctionInvariant
    ppg1.MultiFileInvariant = MultiFileInvariant

    ppg1.MultiFileGeneratingJob = MultiFileGeneratingJob
    ppg1.FileGeneratingJob = FileGeneratingJob
    ppg1.CachedAttributeLoadingJob = CachedAttributeLoadingJob
    ppg1.CachedDataLoadingJob = CachedDataLoadingJob
    ppg1.TempFileGeneratingJob = TempFileGeneratingJob
    ppg1.MultiTempFileGeneratingJob = MultiTempFileGeneratingJob
    ppg1.PlotJob = PlotJob

    ppg1.Job = ppg2.Job  # don't wrap, we need the inheritance
    ppg1.DataLoadingJob = wrap_job(ppg2.DataLoadingJob)
    ppg1.AttributeLoadingJob = wrap_job(ppg2.AttributeLoadingJob)
    ppg1.JobGeneratingJob = wrap_job(ppg2.JobGeneratingJob)

    # unsupported
    for k in (
        "NotebookJob",
        "DependencyInjectionJob",
        "TempFilePlusGeneratingJob",
        "MemMappedDataLoadingJob",
        "FinalJob",
        "CombinedPlotJob",
        "NothingChanged",  # very implementation detail...
    ):
        setattr(ppg1, k, unsupported(k))

    # misc
    ppg1.resource_coordinators = ResourceCoordinators
    ppg1.new_pipegraph = new_pipegraph
    ppg1.run_pipegraph = run_pipegraph
    ppg1.util = util
    ppg1.graph = graph
    ppg1.job = job
    ppg1.JobList = ppg2.JobList
    ppg1.ppg_exceptions = ppg_exceptions
    ppg1.inside_ppg = ppg2.inside_ppg
    ppg1.assert_uniqueness_of_object = ppg2.assert_uniqueness_of_object
    ppg1.testing = ppg2.testing
    ppg1.is_ppg2 = True
    ppg1.testing = ppg2.testing
    ppg1.testing.fixtures = ppg2.testing.fixtures
    # todo: list unpatched...
    new_entries = set(dir(ppg1))
    # this was used to find unported code.
    # for k in set(old_entries).difference(new_entries): # pragma: no cover
    # if not k.startswith("__") and k != "all":
    # warnings.warn(f"not yet ppg1-compatibility layer implemented: {k}")
    patched = True


def unreplace_ppg1():
    """Turn ppg1 compatibility layer off, restoring ppg1
    not that well tested, I suppose...

    """
    global patched
    if not patched:
        return
    for x in dir(ppg1):
        delattr(ppg1, x)
    for k, v in old_entries.items():
        setattr(ppg1, k, v)
    for k, v in old_modules.items():
        sys.modules[k] = v
    ppg1.testing = old_modules["pypipegraph.testing"]
    ppg1.testing.fixtures = old_modules["pypipegraph.testing.fixtures"]
    patched = False


def wrap_job(cls):
    """Adapt for ppg1 api idiosyncracies"""
    return lambda *args, **kwargs: PPG1Adaptor(cls(*args, **kwargs))


class ResourceCoordinators:
    def LocalSystem(max_cores_to_use=ppg2.ALL_CORES, profile=False, interactive=True):
        return (max_cores_to_use, interactive)


class Util:
    @property
    def global_pipegraph(self):
        from . import global_pipegraph

        return global_pipegraph

    @global_pipegraph.setter
    def global_pipegraph(self, value):
        from . import change_global_pipegraph

        change_global_pipegraph(value)

    @staticmethod
    def checksum_file(filename):
        """was used by outside functions"""
        import stat as stat_module
        import hashlib

        file_size = os.stat(filename)[stat_module.ST_SIZE]
        if file_size > 200 * 1024 * 1024:  # pragma: no cover
            print("Taking md5 of large file", filename)
        with open(filename, "rb") as op:
            block_size = 1024**2 * 10
            block = op.read(block_size)
            _hash = hashlib.md5()
            while block:
                _hash.update(block)
                block = op.read(block_size)
            res = _hash.hexdigest()
        return res


def job_or_filename(
    job_or_filename, invariant_class=None
):  # we want to return the wrapped class
    if invariant_class is None:
        invariant_class = FileInvariant
    return ppg2.util.job_or_filename(job_or_filename, invariant_class)


util = Util()
util.job_or_filename = job_or_filename
util.inside_ppg = ppg2.inside_ppg
util.assert_uniqueness_of_object = ppg2.assert_uniqueness_of_object
util.flatten_jobs = ppg2.util.flatten_jobs
util.freeze = ppg2.jobs.ParameterInvariant.freeze


class PPGExceptions:
    pass


ppg_exceptions = PPGExceptions()
for old, new in exception_map.items():
    setattr(ppg_exceptions, old, getattr(ppg2.exceptions, new))


# earlier on, we had a different pickling scheme,
# and that's what the files were called.
if os.path.exists(".pypipegraph_status_robust"):  # old projects keep their filename
    invariant_status_filename_default = ".pypipegraph_status_robust"  # pragma: no cover
elif "/" in sys.argv[0]:  # no script name but an executable?
    invariant_status_filename_default = ".pypipegraph_status_robust"
else:
    # script specific pipegraphs
    invariant_status_filename_default = (
        ".ppg_status_%s" % sys.argv[0]
    )  # pragma: no cover


class Graph:

    invariant_status_filename_default = invariant_status_filename_default


graph = Graph()


class Job:
    _InvariantJob = ppg2.jobs._InvariantMixin
    pass


job = Job()
job.function_to_str = ppg2.FunctionInvariant.function_to_str


class FakeRC:
    @property
    def cores_available(self):
        return ppg2.global_pipegraph.cores


def new_pipegraph(
    resource_coordinator=None,
    quiet=False,
    invariant_status_filename=None,
    dump_graph=True,
    interactive=True,
    cache_folder="cache",
    log_file=None,
    log_level=logging.ERROR,
):
    cores = ppg2.ALL_CORES
    run_mode = ppg2.RunMode.CONSOLE
    if resource_coordinator:
        cores = resource_coordinator[0]
        interactive = resource_coordinator[1]  # rc overrides interactive setting
    if interactive:  # otherwise, we read the one passed into the function
        run_mode = ppg2.RunMode.CONSOLE
    else:
        run_mode = ppg2.RunMode.NONINTERACTIVE
    kwargs = {}
    if invariant_status_filename:
        invariant_status_filename = Path(invariant_status_filename)
        kwargs["log_dir"] = invariant_status_filename / "logs"
        kwargs["error_dir"] = invariant_status_filename / "errors"
        kwargs["history_dir"] = invariant_status_filename / "history"
        kwargs["run_dir"] = invariant_status_filename / "run"
    kwargs["allow_short_filenames"] = False  # as was the default for ppg1
    kwargs["prevent_absolute_paths"] = False  # as was the default for ppg1

    res = ppg2.new(
        cores=cores,
        run_mode=run_mode,
        log_level=log_level,
        cache_dir=Path(cache_folder),
        **kwargs,
    )
    _add_graph_comp(res)
    return res


def _add_graph_comp(graph):
    graph.cache_folder = graph.cache_dir  # ppg1 compatibility
    graph.rc = FakeRC()
    util.global_pipegraph = graph


def run_pipegraph(*args, **kwargs):
    """Run the current global pipegraph"""
    if util.global_pipegraph is None:
        raise ValueError("You need to call new_pipegraph first")
    ppg2.run(**kwargs)


def _ignore_code_changes(job):
    job.depend_on_function = False
    if hasattr(job, "func_invariant"):
        log_job_trace(f"ignoring changes for {job.job_id}")
        util.global_pipegraph.job_dag.remove_edge(job.func_invariant.job_id, job.job_id)

        if hasattr(job.func_invariant, "usage_counter"):
            job.func_invariant.usage_counter -= 1
        if (
            not hasattr(job.func_invariant, "usage_counter")
            or job.func_invariant.usage_counter == 0
        ):
            util.global_pipegraph.job_dag.remove_node(job.func_invariant.job_id)
            for k in job.func_invariant.outputs:
                util.global_pipegraph.job_inputs[job.job_id].remove(k)
            del util.global_pipegraph.jobs[job.func_invariant.job_id]

        del job.func_invariant
    if hasattr(job, "lfg"):
        _ignore_code_changes(job.lfg)


class PPG1AdaptorBase:
    def ignore_code_changes(self):
        _ignore_code_changes(self)

    def use_cores(self, value):
        self.cores_needed = value
        return self

    @property
    def cores_needed(self):
        res = None
        if self.resources == ppg2.Resources.AllCores:
            res = -1
        elif self.resources == ppg2.Resources.Exclusive:
            res = -2
        else:
            res = 1
        return res

    @cores_needed.setter
    def cores_needed(self, value):
        if value == -1 or value > 1:
            self.use_resources(ppg2.Resources.AllCores)
        elif value == -2:
            self.use_resources(ppg2.Resources.Exclusive)
        else:  # elif value == 1:
            self.use_resources(ppg2.Resources.SingleCore)

    def depends_on(self, *args):  # keep the wrapper
        if hasattr(self, "__wrapped__"):
            res = self.__wrapped__.depends_on(*args)
        else:
            super().depends_on(*args)
        return self

    def depends_on_file(self, filename):
        job = FileInvariant(filename)
        self.depends_on(job)
        return ppg2.jobs.DependsOnInvariant(job, self)

    def depends_on_params(self, params):
        job = ParameterInvariant(self.job_id, params)
        self.depends_on(job)
        return ppg2.jobs.DependsOnInvariant(job, self)

    @property
    def filenames(self):
        return self.files

    @property
    def prerequisites(self):
        return self.upstreams


class PPG1Adaptor(wrapt.ObjectProxy, PPG1AdaptorBase):
    pass


class FileInvariant(PPG1AdaptorBase, ppg2.FileInvariant):
    pass


class FunctionInvariant(PPG1AdaptorBase, ppg2.FunctionInvariant):
    pass


class ParameterInvariant(PPG1AdaptorBase, ppg2.ParameterInvariant):
    pass


def assert_ppg_created():
    if not util.global_pipegraph:
        raise ValueError("Must instantiate a pipegraph before creating any Jobs")


def _first_param_empty(signature):
    """Check whether the first argument to this call is
    empty, ie. no with a default value"""
    try:
        first = next((signature.parameters.items()).__iter__())
        return first[1].default == inspect._empty
    except StopIteration:
        return True


def _wrap_func_if_no_output_file_params(function, accept_all_defaults=False):
    sig = inspect.signature(function)
    if len(sig.parameters) == 0 or not _first_param_empty(sig):
        # no or only default parameters = do it oldstyle.
        if not accept_all_defaults and not _first_param_empty(sig):
            raise TypeError(
                f"Could not correctly wrap {function}.\n"
                f"{ppg2.FunctionInvariant.function_to_str(function)}\n"
                "It has default parameter that would have been replaced "
                "with output_filename in ppg1 already. Fix your function arguments"
            )

        def wrapper(of):  # pragma: no cover - runs in spawned process
            function()
            if not isinstance(of, list):
                of = [of]
            for a_filename in of:
                if not a_filename.exists():
                    raise ppg2.exceptions.JobContractError(
                        "%s did not create its file(s) %s %s\n.Cwd: %s"
                        % (
                            a_filename,
                            function.__code__.co_filename,
                            function.__code__.co_firstlineno,
                            os.path.abspath(os.getcwd()),
                        )
                    )

        wrapper.wrapped_function = function
        func = wrapper
    else:
        func = function
    return func


class FileGeneratingJob(PPG1AdaptorBase, ppg2.FileGeneratingJob):
    def __new__(cls, *args, **kwargs):
        obj = ppg2.FileGeneratingJob.__new__(cls, *args, **kwargs)
        return obj

    def __init__(self, output_filename, function, rename_broken=False, empty_ok=False):
        func = _wrap_func_if_no_output_file_params(function)
        super().__init__(output_filename, func, empty_ok=empty_ok)


class MultiFileGeneratingJob(PPG1AdaptorBase, ppg2.MultiFileGeneratingJob):
    def __init__(self, output_filenames, function, rename_broken=False, empty_ok=False):
        func = _wrap_func_if_no_output_file_params(function, accept_all_defaults=True)
        res = super().__init__(output_filenames, func, empty_ok=empty_ok)


class TempFileGeneratingJob(PPG1AdaptorBase, ppg2.TempFileGeneratingJob):
    def __init__(self, output_filename, function, rename_broken=False):
        func = _wrap_func_if_no_output_file_params(function)
        super().__init__(output_filename, func)


class MultiTempFileGeneratingJob(PPG1AdaptorBase, ppg2.MultiTempFileGeneratingJob):
    def __init__(self, output_filenames, function, rename_broken=False):
        func = _wrap_func_if_no_output_file_params(function, accept_all_defaults=True)
        super().__init__(output_filenames, func)


def MultiFileInvariant(filenames):
    # ppg2 already detects when invariants are gained and lost -> no need for
    # special MultiFileInvariant
    res = []
    for f in filenames:
        res.append(ppg2.FileInvariant(f))
    return res


# no one inherits from these, so wrapping in a function is ok, I suppose
# they should have been functions in ppg1.e..


def CachedAttributeLoadingJob(
    cache_filename, target_object, target_attribute, calculating_function
):
    try:
        job = ppg2.CachedAttributeLoadingJob(
            cache_filename, target_object, target_attribute, calculating_function
        )
    except ppg2.JobRedefinitionError as e:
        raise ppg1.JobContractError(str(e))

    return wrap_old_style_lfg_cached_job(job)


def CachedDataLoadingJob(cache_filename, calculating_function, loading_function):
    job = ppg2.CachedDataLoadingJob(
        cache_filename, calculating_function, loading_function
    )
    return wrap_old_style_lfg_cached_job(job)


def PlotJob(
    output_filename,
    calc_function,
    plot_function,
    render_args=None,
    skip_table=False,
    skip_caching=False,
):
    pj = ppg2.PlotJob(
        output_filename,
        calc_function,
        plot_function,
        render_args=render_args,
        cache_calc=not skip_caching,
        create_table=not skip_table,
    )
    res = pj.plot
    if isinstance(pj.cache, ppg2.jobs.CachedJobTuple):
        res.cache_job = wrap_old_style_lfg_cached_job(pj.cache)
    else:
        res.cache_job = pj.cache
    res.table_job = pj.table
    res = PPG1Adaptor(res)

    def depends_on(
        self,
        *other_jobs,
    ):
        # FileGeneratingJob.depends_on(self, other_job)  # just like the cached jobs, the plotting does not depend on the loading of prerequisites
        if res.cache_job is None:
            ppg2.Job.depends_on(self, *other_jobs)
            if self.table_job is not None:
                self.table_job.depends_on(*other_jobs)
        elif (
            hasattr(self, "cache_job") and other_jobs[0] is not self.cache_job
        ):  # activate this after we have added the invariants...
            self.cache_job.depends_on(*other_jobs)
        return self

    res.depends_on = types.MethodType(depends_on, res)

    def ignore_code_changes(self):
        _ignore_code_changes(self)
        if self.cache_job is not None:
            _ignore_code_changes(self.cache_job)
        if self.table_job is not None:
            _ignore_code_changes(self.table_job)

    res.ignore_code_changes = types.MethodType(ignore_code_changes, res)

    return res


def wrap_old_style_lfg_cached_job(job):
    # adapt new style to old style
    if hasattr(job.load, "__wrapped__"):  # pragma: no cover
        res = job.load
        res.lfg = job.calc  # just assume it's a PPG1Adaptor
    else:
        res = PPG1Adaptor(job.load)
        res.lfg = PPG1Adaptor(job.calc)

    def depends_on(self, *args, **kwargs):
        if args and args[0] == self.lfg:  # repeated definition, I suppose
            # must not call self.__wrapped__.depends_on - that's a recursion for some reason?
            ppg2.Job.depends_on(self, *args, **kwargs)
        else:
            self.lfg.depends_on(*args, **kwargs)
        return self

    res.depends_on = depends_on.__get__(res)

    def use_cores(self, cores):
        self.lfg.use_cores(cores)
        return self

    res.use_cores = use_cores.__get__(res)

    return res


def unsupported(name):
    def inner():
        raise NotImplementedError(f"ppg2 no longer offers {name}")

    return inner



#!/usr/bin/env python3
import pypipegraph2 as ppg
import shutil
from pathlib import Path
import os

run_dir = (Path(__file__).parent / "run").absolute()
counter = 0


def setup():
    global counter
    rd = run_dir / str(counter)
    rd.mkdir(parents=True, exist_ok=True)
    os.chdir(rd)
    counter += 1


def simple():
    setup()
    ppg.new()
    jobs = []
    data = {}
    for ii in range(100):

        def load(ii=ii):
            data[ii] = str(ii)

        loader = ppg.DataLoadingJob(f"dl{ii}", load)
        fg = ppg.FileGeneratingJob(
            str(ii),
            lambda of, ii=ii: of.write_text(data[ii]),
            # resources=ppg.Resources.RunsHere,
        )
        fg.depends_on(loader)
        if ii > 10:
            if ii % 2:
                fg.depends_on(jobs[ii % 10])
            else:
                loader.depends_on(jobs[ii % 10])
        jobs.append(fg)
    for i in range(10):
        ppg.run()  # i really want to see the nothing to do parts


def test_simple(benchmark):
    if run_dir.exists():
        shutil.rmtree(run_dir)
    benchmark(simple)


if __name__ == "__main__":
    simple()



import pytest
from pathlib import Path
from .shared import write, force_load
import pypipegraph2 as ppg2
import pypipegraph as ppg1

# to push test coverage...


class TestcompatibilityLayer:
    def test_repeated_apply_unapply(self):
        fg1 = ppg1.FileGeneratingJob
        fg2 = ppg2.FileGeneratingJob
        ppg2.replace_ppg1()
        assert ppg1.FileGeneratingJob is not fg1
        assert ppg1.FileGeneratingJob is not fg2  # actually it is wrapped...
        fg1_to_two = ppg1.FileGeneratingJob
        ppg2.unreplace_ppg1()
        assert ppg1.FileGeneratingJob is fg1
        ppg2.replace_ppg1()
        assert ppg1.FileGeneratingJob is fg1_to_two
        ppg2.replace_ppg1()
        assert ppg1.FileGeneratingJob is fg1_to_two
        ppg2.replace_ppg1()
        assert ppg1.FileGeneratingJob is fg1_to_two
        ppg2.unreplace_ppg1()
        assert ppg1.FileGeneratingJob is fg1
        ppg2.unreplace_ppg1()
        assert ppg1.FileGeneratingJob is fg1


@pytest.mark.usefixtures("ppg1_compatibility_test")
class TestcompatibilityLayerMore:
    def test_use_cores_exclusive(self):
        j = ppg1.FileGeneratingJob("shu", lambda of: of.write_text("j"))
        j.cores_needed = -2
        assert j.resources == ppg2.Resources.Exclusive
        assert j.cores_needed == -2

    def test_fg_did_not_create_its_file(self):
        j = ppg1.FileGeneratingJob("test_file_gen", lambda: 55)  # old school callback
        ppg1.FileGeneratingJob(
            "test_file_gen_does", lambda: Path("test_file_gen_does").write_text("A")
        )  # old school callback
        with pytest.raises(ppg2.JobsFailed):
            ppg2.run()
        assert "did not create" in str(j.exception)

    def test_multifilegenerating_without_arguments(self):
        ppg1.MultiFileGeneratingJob(
            ["out/A", "out/B"], lambda: write("out/A", "A") or write("out/B", "B")
        )
        j2 = ppg1.MultiFileGeneratingJob(["out/C", "out/D"], lambda: 55)
        j3 = ppg1.MultiFileGeneratingJob(["out/G", "out/F"], lambda of: 55)
        with pytest.raises(ppg2.JobsFailed):
            ppg2.run()
        assert "did not create" in str(j2.exception)
        assert "did not create" in str(j3.exception)
        assert Path("out/A").read_text() == "A"
        assert Path("out/B").read_text() == "B"

    def test_temp_file_with_and_without(self):
        a = ppg1.TempFileGeneratingJob("A", lambda: 55)
        b = ppg1.TempFileGeneratingJob("b", lambda of: 55)
        c = ppg1.MultiTempFileGeneratingJob(["C"], lambda: 55)
        d = ppg1.MultiTempFileGeneratingJob(["D"], lambda of: 55)
        force_load(a)
        force_load(b)
        force_load(c)
        force_load(d)
        with pytest.raises(ppg2.JobsFailed):
            ppg2.run()
        assert "did not create" in str(a.exception)
        assert "did not create" in str(b.exception)
        assert "did not create" in str(c.exception)
        assert "did not create" in str(d.exception)

    def test_unsupported(self):
        with pytest.raises(NotImplementedError):
            ppg1.MemMappedDataLoadingJob()

    def test_predecessors(self):
        a = ppg1.TempFileGeneratingJob("A", lambda: 55)
        b = ppg1.TempFileGeneratingJob("b", lambda of: 55)
        b.ignore_code_changes()
        b.depends_on(a)
        assert list(b.prerequisites) == [a]

    def test_depends_on_file_param_returns_wrapped(self):
        a = ppg1.FileGeneratingJob("a", lambda of: write(of))
        Path("input").write_text("hello")
        b = a.depends_on_file("input").invariant
        assert isinstance(b, ppg2.ppg1_compatibility.FileInvariant)
        c = a.depends_on_params("shu").invariant
        assert isinstance(c, ppg2.ppg1_compatibility.ParameterInvariant)

    def test_callback_adaption_with_default_parameters(self):
        def no_args():
            pass

        def all_default_args(a=123, b=234):
            pass

        def new_style(of, a=123, b=234):
            pass

        def new_style2(of):
            pass

        a = ppg1.FileGeneratingJob("a", no_args)
        assert hasattr(
            a.generating_function, "wrapped_function"
        )  # is a wrapped function
        b = ppg1.FileGeneratingJob("b", new_style)
        assert not hasattr(
            b.generating_function, "wrapped_function"
        )  # is a wrapped function
        b1 = ppg1.FileGeneratingJob("b1", new_style)
        assert not hasattr(
            b1.generating_function, "wrapped_function"
        )  # is a wrapped function

        with pytest.raises(TypeError):
            ppg1.FileGeneratingJob("c", all_default_args)
        ppg1.new_pipegraph()
        a = ppg1.MultiFileGeneratingJob(["a"], no_args)
        assert hasattr(
            a.generating_function, "wrapped_function"
        )  # is a wrapped function
        b = ppg1.MultiFileGeneratingJob(["b"], new_style)
        assert not hasattr(
            b.generating_function, "wrapped_function"
        )  # is a wrapped function
        ppg1.MultiFileGeneratingJob(
            ["c"], all_default_args
        )  # mfg never passed [output_files]
        assert hasattr(
            a.generating_function, "wrapped_function"
        )  # is a wrapped function

    def test_rc_cores_available(self):
        assert (
            ppg1.util.global_pipegraph.rc.cores_available == ppg2.global_pipegraph.cores
        )

    def test_ignore_code_changes_changes_both_dag_and_jobs(self):
        a = ppg1.FileGeneratingJob("a", lambda of: of.write_text("a"))
        assert len(ppg2.global_pipegraph.jobs) == 2  # the fg, and teh FI
        assert len(ppg2.global_pipegraph.job_dag) == 2  # the fg, and teh FI
        a.ignore_code_changes()
        assert len(ppg2.global_pipegraph.jobs) == 1  # the fg, and teh FI
        assert len(ppg2.global_pipegraph.job_dag) == 1  # the fg, and teh FI



# -*- coding: utf-8 -*-
__version__ = "3.0.6"

from pathlib import Path
import logging
import contextlib
from .graph import PyPipeGraph, ALL_CORES
from .jobs import (
    FileGeneratingJob,
    MultiFileGeneratingJob,
    TempFileGeneratingJob,
    MultiTempFileGeneratingJob,
    DataLoadingJob,
    ValuePlusHash,
    AttributeLoadingJob,
    CachedDataLoadingJob,
    CachedAttributeLoadingJob,
    PlotJob,
    FunctionInvariant,
    FileInvariant,
    ParameterInvariant,
    JobGeneratingJob,
    Job,
    JobList,
    SharedMultiFileGeneratingJob,
    NotebookInvariant,
    NotebookJob
)
from .exceptions import (
    PPGException,
    NotADag,
    FatalGraphException,
    JobOutputConflict,
    JobContractError,
    JobDied,
    JobRedefinitionError,
    RunFailed,
    RunFailedInternally,
    JobsFailed,
    JobError,
    HistoryLoadingFailed,
    JobEvaluationFailed,
)
from .enums import Resources, RunMode
from . import util
from .util import assert_uniqueness_of_object
from .pypipegraph2 import enable_logging as enable_rust_logging

reuse_last_or_default = object()
default = object()

_last_new_arguments = {}
_ppg1_compatibility_mode = False


def _last_or_default(name, value, default_value):
    if value is default:  # pragma: no cover
        result = default_value
    elif value is reuse_last_or_default:
        result = _last_new_arguments.get(name, default_value)
    else:
        result = value
    _last_new_arguments[name] = result
    return result


def new(
    cores=reuse_last_or_default,
    run_mode=reuse_last_or_default,
    log_dir=reuse_last_or_default,
    error_dir=reuse_last_or_default,
    history_dir=reuse_last_or_default,
    run_dir=reuse_last_or_default,
    log_level=reuse_last_or_default,
    allow_short_filenames=reuse_last_or_default,
    log_retention=reuse_last_or_default,
    cache_dir=reuse_last_or_default,
    prevent_absolute_paths=reuse_last_or_default,
    report_done_filter=reuse_last_or_default,
):
    """create a new pipegraph.
    You may pass reuse_last_or_default to all values
    to reuse the last value,
    or default to use the true default

    (or load defaults)


    Log retention is how many old logs (+ the current one) we
    keep.
    """
    global global_pipegraph
    locs = locals()
    arguments = {
        name: _last_or_default(name, locs[name], default_arg)
        for name, default_arg in [
            ("cores", ALL_CORES),
            ("log_dir", Path(".ppg/logs")),
            ("error_dir", Path(".ppg/errors")),
            ("history_dir", Path(".ppg/history")),
            ("run_dir", Path(".ppg/run")),
            ("log_level", logging.DEBUG),  # that's the one for the log file
            ("allow_short_filenames", False),
            ("run_mode", RunMode.CONSOLE),
            ("log_retention", 3),
            ("cache_dir", Path("cache")),
            ("prevent_absolute_paths", True),
            ("report_done_filter", 1),
        ]
    }
    util.do_jobtrace_log = arguments["log_level"] <= 6
    # if arguments['run_mode'] != RunMode.NONINTERACTIVE:
    # raise ValueError()
    global_pipegraph = PyPipeGraph(**arguments)

    if _ppg1_compatibility_mode:
        from . import ppg1_compatibility

        ppg1_compatibility._add_graph_comp(global_pipegraph)

    return global_pipegraph


global_pipegraph = None


def change_global_pipegraph(value):
    """Helper to swap out the global pipegraph from ppg1-compability.util"""
    global global_pipegraph
    global_pipegraph = value


def run(
    print_failures=True,
    raise_on_job_error=True,
    event_timeout=1,
    dump_graphml=None,
):
    if global_pipegraph is None:
        raise ValueError("Must instantiate a pipegraph before you can run it.")

    return global_pipegraph.run(
        print_failures=print_failures,
        raise_on_job_error=raise_on_job_error,
        event_timeout=event_timeout,
        dump_graphml=dump_graphml,
    )


def inside_ppg():
    return global_pipegraph is not None


@contextlib.contextmanager
def _with_changed_global_pipegraph(new):
    global global_pipegraph
    old = global_pipegraph
    try:
        global_pipegraph = new
        yield new
    finally:
        global_pipegraph = old


def replace_ppg1():
    """Turn all ppg1 references into actual ppg2
    objects.
    Best effort, but the commonly used API should be well supported.
    Try to do this before anything imports ppg1.


    One notably exception is in cores_needed/use_cores,
    where ppg2 only supports 1/almost_all/all, and ppg1 also supported
    a number of cores > 1. This get's convert to almost_all (-1 in ppg1 parlance)

    Also jobs often throw TypeError, instead of ValueError if you pass the arguments
    in the wrong order. This shouldn't affect any working code though.
    """
    from . import ppg1_compatibility

    global _ppg1_compatibility_mode
    _ppg1_compatibility_mode = True
    ppg1_compatibility.replace_ppg1()


def unreplace_ppg1():
    """undo replace_ppg1."""
    from . import ppg1_compatibility

    global _ppg1_compatibility_mode
    _ppg1_compatibility_mode = False

    ppg1_compatibility.unreplace_ppg1()


__all__ = [
    "new",
    "run",
    "Job",
    "JobList",
    "FileGeneratingJob",
    "MultiFileGeneratingJob",
    "TempFileGeneratingJob",
    "MultiTempFileGeneratingJob",
    "MultiFileGeneratingJob",
    "DataLoadingJob",
    "ValuePlusHash",
    "AttributeLoadingJob",
    "CachedDataLoadingJob",
    "CachedAttributeLoadingJob",
    "PlotJob",
    "FunctionInvariant",
    "FileInvariant",
    "ParameterInvariant",
    "NotebookInvariant",
    "NotebookJob",
    "JobGeneratingJob",
    "SharedMultiFileGeneratingJob",
    "PPGException",
    "NotADag",
    "FatalGraphException",
    "JobOutputConflict",
    "JobContractError",
    "JobDied",
    "JobRedefinitionError",
    "RunFailed",
    "RunFailedInternally",
    "JobsFailed",
    "JobError",
    "HistoryLoadingFailed",
    "JobEvaluationFailed",
    "Resources",
    "RunMode",
    "replace_ppg1",
    "unreplace_ppg1",
    "inside_ppg",
    "assert_uniqueness_of_object",
    "enable_rust_logging",
]



import shutil
import sys
import pytest

from pathlib import Path
import pypipegraph2 as ppg
from .shared import write, counter

# fmt: off
# has to be here for the indent to be the same...
# since the test case also change python versions


def _test_rewriting_history_do_a(ofs):
    counter('a')
    for of in ofs:
        of.write_text(of.name)
# fmt: on


@pytest.mark.usefixtures("ppg2_per_test")
class TestRustConversion:
    def test_rewriting_history(self):
        # not for python <  3.8
        if sys.version_info[:2] < (3, 8):
            return
        input = Path(__file__).parent / "old_history_for_conversion_test.gz"
        dir = ppg.global_pipegraph.get_history_filename().parent
        dir.mkdir(parents=True)
        old_history_file = dir / "ppg_history.gz"
        shutil.copy(input, old_history_file)

        write("A", "A")
        write("C", "C")
        write("B", "B")

        jobA = ppg.MultiFileGeneratingJob(
            ["A", "C"],
            _test_rewriting_history_do_a,
            resources=ppg.Resources.AllCores,
            depend_on_function=False,
        )
        # I have to make the func invariant here myself.
        # the example was dumped from a run.py
        # where funcinvariant name = 'do_a'
        # but here were inside a func, so it's 'FITestRustConversion.test_rewriting_history.<locals>.do_a'
        # jobA.depends_on(ppg.FunctionInvariant("A:::C", do_a))
        jobA.depends_on(ppg.FunctionInvariant("do_a", _test_rewriting_history_do_a))

        jobB = ppg.FileGeneratingJob(
            "B",
            lambda of: counter("b") and of.write_text(of.name),
            resources=ppg.Resources.AllCores,
        )
        jobB.depends_on("A")
        # ppg.pypipegraph2.enable_logging()
        ppg.run()
        assert not old_history_file.exists()  # g got renamed

        assert not Path("a").exists()
        assert not Path("b").exists()



# showcase how a nested exception will look like
import pypipegraph2 as ppg
from pathlib import Path
import os
import shutil

p = Path("run/nested_exceptions")
if p.exists():
    shutil.rmtree(p)
p.mkdir(exist_ok=True, parents=True)
os.chdir(p)


def a():
    try:
        raise ValueError()
    except ValueError as e:
        raise KeyError() from e


ppg.DataLoadingJob("a", a)
try:
    ppg.run()
except ppg.RunFailed:
    pass


try:
    raise ValueError()
except ValueError:
    raise KeyError()



from typing import Optional, Union, Dict
import gzip
import threading
import json
import logging
import shutil
import collections
import os
import sys
import pickle
import signal
import networkx
import subprocess
import time
import datetime
from pathlib import Path
from loguru import logger

from . import exceptions
from .runner import Runner
from .enums import RunMode, JobOutcome
from .exceptions import JobsFailed, _RunAgain
from .util import (
    CPUs,
    log_info,
    log_error,
    log_warning,
    # log_debug,
    log_trace,
)
from . import util
from rich.logging import RichHandler
from rich.console import Console


logger.level("JT", no=6, color="<yellow>", icon="🐍")
logger.level("INFO", color="")
# if "pytest" in sys.modules:  # pragma: no branch
# log_out = sys.stderr
# else:  # pragma: no cover
# log_out = RichHandler(markup=True, console=console)
# logger.add(
# sink=log_out,
# format="{elapsed} {message}",
# level=logging.INFO,
# )

time_format = "%Y-%m-%d_%H-%M-%S"

start_cwd = Path(".").absolute()


class ALL_CORES:
    pass


class PyPipeGraph:
    history_dir: Optional[Path]
    log_dir: Optional[Path]
    log_level: int
    running: bool

    def __init__(
        self,
        cores: Union[int, ALL_CORES],
        log_dir: Optional[Path],
        error_dir: Optional[Path],
        history_dir: Path,
        run_dir: Path,
        cache_dir: Path,
        log_level: int,
        run_mode: RunMode,
        paths: Optional[Dict[str, Union[Path, str]]] = None,
        allow_short_filenames=False,
        log_retention=None,
        prevent_absolute_paths=True,
        report_done_filter=1,
    ):
        if cores is ALL_CORES:
            self.cores = CPUs()
        else:
            self.cores = int(cores)
        if log_dir:
            self.log_dir = Path(log_dir)
        else:
            self.log_dir = None
        if error_dir:
            self.error_dir = error_dir
        else:
            self.error_dir = None
        self.history_dir = Path(history_dir)
        self.run_dir = Path(run_dir)
        self.log_level = log_level
        self.log_retention = log_retention
        # self.paths = {k: Path(v) for (k, v) in paths} if paths else {}
        self.run_mode = run_mode
        self.jobs = {}  # the job objects, by id
        self.job_dag = (
            networkx.DiGraph()
        )  # a graph. Nodes: job_ids, edges -> must be done before
        self.job_inputs = collections.defaultdict(
            set
        )  # necessary inputs (ie. outputs of other jobs)
        self.outputs_to_job_ids = (
            {}
        )  # so we can find the job that generates an output: todo: should be outputs_to_job_id or?
        self.run_id = 0
        self._test_failing_outside_of_job = False
        self.allow_short_filenames = allow_short_filenames
        if cache_dir:
            self.cache_dir = Path(cache_dir)
            self.cache_dir.mkdir(exist_ok=True, parents=True)
        else:
            self.cache_dir = None
        self.cache_folder = self.cache_dir  # todo: change all occurances?
        self.running = False
        self.prevent_absolute_paths = prevent_absolute_paths
        self._debug_allow_ctrl_c = False  # see examples/abort_when_stalled.py
        self.next_job_number = 0
        self.next_job_number_lock = threading.Lock()
        self._path_cache = {}
        self.report_done_filter = report_done_filter
        self.func_cache = {}
        self.dir_absolute = Path(".").absolute()
        self._jobs_do_dump_subgraph_debug = False
        self._jobs_to_prune_unrelated = False
        self._jobs_to_prune_but = False

    def run(
        self,
        print_failures: bool = True,
        raise_on_job_error=True,
        event_timeout=5,
        dump_graphml=False,
    ) -> Dict[str, JobOutcome]:
        """Run the complete pypipegraph"""
        try:
            return self._run(
                print_failures,
                raise_on_job_error,
                event_timeout,
                None,
                dump_graphml=dump_graphml,
            )
        except JobsFailed as e:  # shorten the traceback considerably!
            raise JobsFailed(e.args[0], exceptions=e.exceptions)

    def _run(
        self,
        print_failures: bool = True,
        raise_on_job_error=True,
        event_timeout=5,
        focus_on_these_jobs=None,
        dump_graphml=False,
    ) -> Dict[str, JobOutcome]:
        """Run the jobgraph - possibly focusing on a subset of jobs (ie. ignoring
        anything that's not necessary to calculate them - activated by calling a Job
        """
        ts = str(
            time.time()
        )  # include subsecond in log names - usefull for the testing, I suppose.
        ts = ts[ts.rfind(".") :]
        self.time_str = datetime.datetime.now().strftime(time_format) + ts
        if not networkx.algorithms.is_directed_acyclic_graph(self.job_dag):
            print(networkx.readwrite.json_graph.node_link_data(self.job_dag))
            raise exceptions.NotADag()
        else:
            # print(networkx.readwrite.json_graph.node_link_data(self.job_dag))
            pass
        start_time = time.time()
        self._resolve_dependency_callbacks()
        self.running = True  # must happen after dependency callbacks
        if self.error_dir:
            self._cleanup_errors()
            (self.error_dir / self.time_str).mkdir(exist_ok=True, parents=True)
            self._link_errors()
        log_id = None
        if self.log_dir:
            self._cleanup_logs()
            self.log_dir.mkdir(exist_ok=True, parents=True)
            fn = Path(sys.argv[0]).name
            self.log_file = self.log_dir / f"{fn}-{self.time_str}.log"
            logger.remove()  # no default logging
            logger.add(
                open(self.log_file, "w"), level=min(self.log_level, logging.DEBUG)
            )
            if False:
                logger.add(
                    RichHandler(
                        markup=False,
                        console=Console(
                            file=open(self.log_dir / f"{fn}-{self.time_str}.log", "w"),
                            width=120,  #
                        ),
                    ),
                    level=self.log_level,
                )
            # if "pytest" in sys.modules:  # pragma: no branch
            log_id = logger.add(
                sink=sys.stdout,
                level=logging.INFO
                if not util.do_jobtrace_log
                else 6,  # don't spam stdout
                format=(
                    "\r  <blue>{elapsed}s</blue> <bold>|</bold> <level>{message}</level>"
                    if not util.do_jobtrace_log
                    else "<blue>{elapsed}s</blue> | <level>{level.icon}</level> <bold>|</bold>{file:8.8}:{line:4} <level>{message}</level>"
                ),
            )  # pragma: no cover

            self._link_logs()

            log_info(
                f"Run is go {threading.get_ident()} pid: {os.getpid()}, run_id {self.run_id}, log_level = {self.log_level}"
            )
        self.history_dir.mkdir(exist_ok=True, parents=True)
        self.run_dir.mkdir(exist_ok=True, parents=True)
        self.do_raise = []
        self._restart_afterwards = False

        # we want logs, so we do tihs 'late'.
        if self._jobs_to_prune_unrelated:
            self._prune_unrelated()
        if self._jobs_to_prune_but:
            self._prune_but()
        ok = False

        try:
            result = None
            self._install_signals()
            try:
                history = self._load_history()
            except Exception as e:
                log_error(f"{e}")
                raise exceptions.HistoryLoadingFailed(e)
            max_runs = 5
            jobs_already_run = set()
            final_result = {}
            aborted = False
            try:
                while True:
                    max_runs -= 1
                    if max_runs == 0:  # pragma: no cover
                        raise ValueError(
                            "Maximum graph-generating-jobs recursion depth exceeded"
                        )
                    do_break = False
                    job_count = len(self.job_dag)
                    try:
                        print("Run history")
                        self.runner = Runner(
                            self,
                            history,
                            event_timeout,
                            focus_on_these_jobs,
                            jobs_already_run,
                            dump_graphml,
                            self.run_id,
                            self._jobs_do_dump_subgraph_debug,
                        )
                        result, new_history = self.runner.run(
                            history, result, print_failures=print_failures
                        )
                        aborted = self.runner.aborted
                        del self.runner
                        self.run_id += 1
                        do_break = True
                    except _RunAgain as e:
                        log_info("Jobs created - running again")
                        result, new_history = e.args[0]
                    self._log_runtimes(result, start_time)
                    # assert len(result) == job_count # does not account for cleanup jobs...
                    # leave out the cleanup jobs added virtually by the run
                    jobs_already_run.update(
                        (k for k in result.keys() if k in self.jobs)
                    )
                    log_info(f"Result len {len(result)}")
                    for k, v in result.items():
                        if (
                            not k in final_result
                            or final_result[k].outcome != JobOutcome.Failed
                        ):
                            final_result[k] = v
                    # final_result.update(result)
                    if history != new_history:
                        self._save_history(new_history)
                    history = new_history

                    if do_break:
                        break
                    # final_result.update(result)
            except KeyboardInterrupt:
                self.do_raise.append(KeyboardInterrupt())
            del result
            print("leave history")
            log_info(f"Left graph loop {len(final_result)}")
            jobs_failed = False
            for job_id, job_state in final_result.items():
                if job_state.outcome == JobOutcome.Failed:
                    self.do_raise.append(job_state.payload)
                    jobs_failed = True
            self.last_run_result = final_result
            if raise_on_job_error and self.do_raise and not self._restart_afterwards:
                if jobs_failed:
                    raise exceptions.JobsFailed(
                        "At least one job failed", self.do_raise
                    )
                else:
                    raise exceptions.RunFailedInternally()
            elif not raise_on_job_error and self.do_raise:
                log_error("At least one job failed")
            if aborted:
                raise KeyboardInterrupt("Run aborted")  # pragma: no cover
            ok = True
            return final_result
        except Exception as e:
            print(e)
            raise
        finally:
            if ok:
                log_info("Run is done")
            else:
                log_info("Run is done - with failure")
            log_info("")
            self._restore_signals()
            if log_id is not None:
                logger.remove(log_id)
            self.running = False
            if (
                self._restart_afterwards
            ):  # pragma: no cover - todo: test with interactive
                log_info("Restart again issued - restarting via subprocess.check_call")
                subprocess.check_call([sys.executable] + sys.argv, cwd=start_cwd)

    def run_for_these(self, jobs):
        """Run graph for just these jobs (and their upstreams), ignoring everything else"""
        if not isinstance(jobs, list):
            jobs = [jobs]
        return self._run(
            print_failures=True, raise_on_job_error=True, focus_on_these_jobs=jobs
        )

    def _cleanup_logs(self):
        """Clean up old logs and drop a 'latest' symlink"""
        if not self.log_dir or self.log_retention is None:  # pragma: no cover
            return
        fn = Path(sys.argv[0]).name
        pattern = f"{fn}-*.log"
        files = sorted(self.log_dir.glob(pattern))
        if len(files) > self.log_retention:
            remove = files[: -self.log_retention]
            for f in remove:
                os.unlink(f)

    def _link_latest(self, dir, pattern, latest_name, target_is_directory):
        link_name = dir / latest_name
        if link_name.exists() or link_name.is_symlink():
            # print("unlinking", link_name)
            link_name.unlink()
        # else:
        # print("not found", link_name)

        files = sorted(dir.glob(pattern))
        if files:
            link_name.symlink_to(
                files[-1].name, target_is_directory=target_is_directory
            )

    def _link_logs(self):
        fn = Path(sys.argv[0]).name
        self._link_latest(self.log_dir, f"{fn}-*.log", "latest", False)

    def _cleanup_errors(self):
        """Cleanup old errors and drop a 'latest' symlink"""
        if not self.error_dir or self.log_retention is None:  # pragma: no cover
            return
        err_dirs = sorted(
            [
                x
                for x in (self.error_dir / self.time_str).parent.glob("*")
                if x.is_dir() and not x.is_symlink()
            ]
        )
        if len(err_dirs) > self.log_retention:
            remove = err_dirs[: -self.log_retention]
            for f in remove:
                shutil.rmtree(f)

    def _link_errors(self):
        self._link_latest(self.error_dir, "*", "latest", True)

    def _log_runtimes(self, job_results, run_start_time):
        """Log the runtimes to a file (ever growing. But only runtimes over a threshold)"""
        if self.log_dir:
            rt_file = self.log_dir / "runtimes.tsv"
            lines = []
            if not rt_file.exists():
                lines.append("jobid\trun_start_time\truntime_s")
            for job_id, job_result in job_results.items():  # pragma: no branch
                if job_result.outcome is JobOutcome.Success:
                    if job_result.run_time >= 1:
                        lines.append(
                            f"{job_id}\t{int(run_start_time)}\t{job_result.run_time:.2f}"
                        )
            with open(rt_file, "a+") as op:
                op.write("\n".join(lines))

    def get_history_filename(self):
        """where do we store the graph's history?"""
        # we by default share the history file
        # if it's the same history dir, it's the same project
        # and you'd retrigger the calculations too often otherwise
        return self.history_dir / "ppg_history.2.gz"  # don't end on .py

    def _load_history(self):
        log_trace("_load_history")
        fn = self.get_history_filename()
        self._convert_old_history()
        history = {}
        if fn.exists():
            log_trace("Historical existed")
            with gzip.GzipFile(fn, "rb") as op:
                history = json.loads(op.read().decode("utf-8"))

        log_info(f"Loaded {len(history)} history entries")
        return history

    def _save_history(self, historical):
        log_trace("_save_history")
        fn = self.get_history_filename()
        if Path(fn).exists():
            fn.rename(fn.with_suffix(fn.suffix + ".backup"))
        raise_keyboard_interrupt = False

        try_again = True
        while try_again:
            with gzip.GzipFile(fn, "wb") as op:
                try:
                    op.write(json.dumps(historical, indent=2).encode("utf-8"))
                    if self._test_failing_outside_of_job:  # for unit test
                        self._test_failing_outside_of_job = False
                        log_error("keyboard interrupt raised in loop")
                        raise KeyboardInterrupt("simulated")
                    try_again = False
                except KeyboardInterrupt:
                    raise_keyboard_interrupt = True
        if raise_keyboard_interrupt:
            log_error("Keyboard interrupt during history dumping")
            raise KeyboardInterrupt()

    def _convert_old_history(self):
        """Convert pre-rust history,
        by assuming that it's the same graph, and you changed *nothing* else in the upgrade
        """
        old_filename = self.history_dir / "ppg_history.gz"
        new_filename = self.get_history_filename()
        if not old_filename.exists():
            return

        def failed(reason):
            sys.stderr.write("Could not convert history.\n")
            sys.stderr.write(f"Reason: {reason}.\n\n")
            sys.stderr.write(
                f"Safe choice: Remove {self.history_dir}, rebuild everything\n"
            )
            sys.stderr.write(
                f"Unsafe choice: Remove {self.history_dir}/ppg_history.gz, keeping ppg_history.2.gz.\n"
            )
            sys.stderr.write(
                "The old history had no knowledge of the set of inputs of a job\n"
            )
            sys.stderr.write(
                "So for the new one we must assume that those did not change.\n"
            )

            sys.stderr.write("And we can't do that automatically.\n")
            sys.stderr.write("\nAborting\n")
            sys.exit(1)

        if old_filename.exists() and new_filename.exists():
            raise ValueError(
                "Both new and old history existed. Presumably a failed conversion job. Aborting"
            )
        log_warning("Converting old history to new style")

        old = self._load_old_history(old_filename)  # todo
        new = {}
        # for job_id in self.jobs.keys():
        # new[job_id] = json.dumps(old[job_id], indent=2)

        for job_id, job in self.jobs.items():
            upstreams = job.upstreams
            # incoming_edges = sorted([x.job_id for x in upstreams])
            # "\n".join(incoming_edges)
            new[job_id + "!!!"] = Runner.get_job_inputs_str(self, job_id)
            new[job_id] = json.dumps(old[job_id][1])  # 1, that's the output hash.

            for input_name in self.job_inputs[job_id]:
                upstream_job_id = self.outputs_to_job_ids[input_name]
                key = f"{upstream_job_id}!!!{job_id}"
                new[key] = json.dumps(
                    {input_name: old[job_id][0][input_name]}
                )  # 0 - that's the input hashs.

        for k, v in new.items():
            if not isinstance(v, str):
                raise ValueError(k, v)
        self._save_history(new)

        keys = set(self.jobs.keys())
        if len(old) != len(keys):
            missing = set(old.keys()).difference(keys)
            gained = set(keys).difference(old.keys())
            return failed(
                f"Number of jobs changed. Was {len(old)} is now {len(keys)}. Missing {missing}. Gained: {gained}"
            )
        if keys != set(old):
            return failed("Job names changed")

        old_filename.rename(old_filename.with_suffix(".1.gz.converted_to_2_format"))

    @staticmethod
    def _load_old_history(path):
        with gzip.GzipFile(Path(path)) as of:
            old = {}
            try:
                while True:
                    key = pickle.load(of)
                    value = pickle.load(of)
                    if key in old:
                        raise KeyError()
                    old[key] = value
            except EOFError:
                pass
        return old

    def _resolve_dependency_callbacks(self):
        """jobs may depend on functions that return their actual dependencies.
        This resolves them
        """
        # we need this copy,
        # for the callbacks may create jobs
        # so we can't simply iterate over the jobs.values()
        with_callback = [j for j in self.jobs.values() if j.dependency_callbacks]
        # log_info(f"with callbacks {[j.job_id for j in with_callback]}")
        if not with_callback:
            return
        for j in with_callback:
            dc = j.dependency_callbacks
            j.dependency_callbacks = (
                []
            )  # must reset before run, might add new ones, right?
            for c in dc:
                # log_info(f"{j.job_id}, {c}")
                j.depends_on(c())
        self._resolve_dependency_callbacks()  # nested?

    def _install_signals(self):
        """make sure we don't crash just because the user logged of.
        Also blocks CTRl-c in console, and transaltes into save shutdown otherwise.
        """
        log_trace("_install_signals")

        def hup(*args, **kwargs):  # pragma: no cover
            log_warning("user logged off - continuing run")

        def sigint(*args, **kwargs):
            if self.run_mode is (RunMode.CONSOLE):
                if self._debug_allow_ctrl_c == "abort":  # pragma: no cover
                    log_info("CTRL-C from debug - calling interactive abort")
                    self.runner.interactive._cmd_abort(
                        None
                    )  # for testing the abort facility.
                elif self._debug_allow_ctrl_c == "stop":  # pragma: no cover
                    log_info("CTRL-C from debug - calling interactive stop")
                    self.runner.interactive._cmd_default()
                    self.runner.interactive._cmd_stop(
                        None
                    )  # for testing the abort facility.
                elif self._debug_allow_ctrl_c == "stop&abort":  # pragma: no cover
                    log_info("CTRL-C from debug - calling interactive stop")
                    self.runner.interactive._cmd_stop(
                        None
                    )  # for testing the abort facility.
                    self._debug_allow_ctrl_c = "abort"

                else:
                    log_info("CTRL-C has been disabled. Type 'abort<CR>' to abort")
                # TODO remove
            else:  # pragma: no cover - todo: interactive
                log_info("CTRL-C received. Killing all running jobs.")
                if hasattr(self, "runner"):
                    print("calling abort")
                    self.runner.abort()

        if self.run_mode is (RunMode.CONSOLE):
            self._old_signal_hup = signal.signal(signal.SIGHUP, hup)
        # if self.run_mode in (RunMode.CONSOLE, RunMode.NOTEBOOK):
        # we always steal ctrl c
        self._old_signal_int = signal.signal(signal.SIGINT, sigint)

    def _restore_signals(self):
        """Restore signals to pre-run values"""
        log_trace("_restore_signals")
        if hasattr(self, "_old_signal_hup"):  # pragma: no branch
            signal.signal(signal.SIGHUP, self._old_signal_hup)
        if hasattr(self, "_old_signal_int"):  # pragma: no branch
            signal.signal(signal.SIGINT, self._old_signal_int)

    def add(self, job):
        """Add a job.
        Automatically called when a Job() is created
        """

        for output in job.outputs:
            if output in self.outputs_to_job_ids:
                # already being done somewhere else
                if self.outputs_to_job_ids[output] == job.job_id:
                    # but it is in essence the same same job
                    pass  # we replace the job, keeping upstreams/downstream edges
                else:
                    # if self.run_mode != RunMode.NOTEBOOK: todo: accept in notebooks by removing the other  jobs and warning.
                    raise exceptions.JobOutputConflict(
                        job, self.jobs[self.outputs_to_job_ids[output]]
                    )
            self.outputs_to_job_ids[
                output
            ] = job.job_id  # todo: seperate this into two dicts?
        # we use job numbers during run
        # to keep output files unique etc.

        # this does not happen - jobs are deduped by job_id, and it's up to the job
        # to verify that it's not an invalid redefinition
        # if job.job_id in self.jobs and self.jobs[job.job_id] is not job:
        # if self.run_mode.is_strict():
        # raise ValueError(
        # "Added new job in place of old not supported in run_mode == strict"
        # f"new job: {job} id: {id(job)}",
        # f"old job: {self.jobs[job.job_id]} id: {id(self.jobs[job.job_id])}",
        # )
        if not self.running:  # one core., no locking
            job.job_number = self.next_job_number
            self.next_job_number += 1
        else:  # multiple jobGeneratingJobs might be creating jobs at the same time.
            with self.next_job_number_lock:
                job.job_number = self.next_job_number
                self.next_job_number += 1
        self.jobs[job.job_id] = job
        self.job_dag.add_node(job.job_id)
        # assert len(self.jobs) == len(self.job_dag) - we verify this when running

    def find_job_from_file(self, filename):
        if isinstance(filename, Path):
            filename = str(filename)
        return self.jobs[self.outputs_to_job_ids[filename]]

    def find_job(self, job_or_id_or_path):
        from .jobs import Job

        if isinstance(job_or_id_or_path, (str, Path)):
            return self.find_job_from_file(job_or_id_or_path)
        elif isinstance(job_or_id_or_path, Job):
            return job_or_id_or_path
        else:
            raise TypeError("job was not a Job nor a job_id")

    def add_edge(self, upstream_job, downstream_job):
        """Declare a dependency between jobs

        Implementation note:
        While we connect the DAG based on the Jobs,
        we calculate invalidation based on the job_inputs (see Job.depends_on),
        allowing a MultiFileGeneratingJob to (optionally) only invalidate some of it's downstreams.
        """
        if not upstream_job.job_id in self.jobs:
            raise KeyError(f"{upstream_job} not in this graph. Call job.readd() first")
        if not downstream_job.job_id in self.jobs:
            raise KeyError(
                f"{downstream_job} not in this graph. Call job.readd() first"
            )

        self.job_dag.add_edge(upstream_job.job_id, downstream_job.job_id)

    def has_edge(self, upstream_job, downstream_job):
        """Does this edge already exist?"""
        if not isinstance(upstream_job, str):
            upstream_job_id = upstream_job.job_id
        else:
            upstream_job_id = upstream_job
        if not isinstance(downstream_job, str):
            downstream_job_id = downstream_job.job_id
        else:
            downstream_job_id = downstream_job
        return self.job_dag.has_edge(upstream_job_id, downstream_job_id)

    def restart_afterwards(self):
        """Restart the whole python program afterwards?
        Used by the interactive console
        """
        self._restart_afterwards = True  # pragma: no cover - todo: interactive

    def dump_subgraph_for_debug(self, jobs):  # pragma: no cover
        """Write a subgraph_debug.py
        with a faked-out version of this graph.
        See Job.dump_subgraph_for_debug for details"""
        if jobs:
            jall = list(jobs)
        else:
            jall = list(self.jobs.keys())
        j1 = self.jobs[jall[0]]
        j1.dump_subgraph_for_debug(jall)

    def dump_subgraph_for_debug_at_run(self, job_ids):  # pragma: no cover
        """Write a subgraph_debug.py
        with a faked-out version of this graph.
        See Job.dump_subgraph_for_debug for details.

        This version dumps *after* the cleanup jobs
        and so on have been assembled (ie. when run is called)

        """
        if job_ids:
            jall = list(job_ids)
        else:
            jall = list(self.jobs.keys())
        # j1 = self.jobs[jall[0]]
        # j1.dump_subgraph_for_debug(jall)
        self._jobs_do_dump_subgraph_debug = jall

    def prune_all_unrelated_to(self, job_ids):
        """Remove all jobs that are not related to the given jobs.
        This takes effect *after* the cleanup jobs
        and so on have been assembled (ie. when run is called)

        """
        self._jobs_to_prune_unrelated = job_ids

    def _prune_unrelated(self):
        keep = []
        # graph_undirected = self.job_dag.to_undirected()
        # for job_id in self._jobs_to_prune_unrelated:
        #     keep.extend(networkx.node_connected_component(graph_undirected, job_id))
        for job_id in self._jobs_to_prune_unrelated:
            keep.extend(networkx.ancestors(self.job_dag, job_id))
            keep.extend(networkx.descendants(self.job_dag, job_id))

        keep = set(keep)
        log_info(f"keeping {len(keep)} jobs")
        for k in keep:
            log_info(f"keeping {k}")
        if len(keep) == len(self.jobs):
            log_warning("pruning all unrelated jobs, but *everything's connected*")
        to_remove = set(self.jobs.keys()) - keep
        for k in to_remove:
            log_info(f"pruning unrelated job {k}")
            self.jobs[k].prune()

    def prune_all_but(self, job_ids):
        self._jobs_to_prune_but = job_ids

    def _prune_but(self):
        """prune after _resolve_dependency_callback"""
        keep = set(self._jobs_to_prune_but)
        for j in self.jobs:
            if not j in keep:
                self.jobs[j].prune()



# check if we can stop whilst stalling
# only one job should finish.
import pypipegraph2 as ppg
import time
from pathlib import Path
import logging
import os
import shutil
import psutil
import subprocess

p = Path("run/stall")
if p.exists():
    shutil.rmtree(p)
p.mkdir(exist_ok=True, parents=True)
os.chdir(p)

time_before_abort = 1
timeout = 10
jobcount = 5
ppg.new(log_level=logging.INFO, cores=5)
# tell the ppg to accept the SIGINT we're actually sending
ppg.global_pipegraph._debug_allow_ctrl_c = "stop&abort"


def all_cores(ii):
    def inner(of, ii=ii):
        of.write_text(str(time.time()))
        proc = psutil.Process()
        parent = proc.parent()
        if ii == 0:  # only the first guy kills us
            time.sleep(time_before_abort)
            subprocess.check_call(["kill", "--signal", "SIGINT", str(parent.pid)])
            time.sleep(1)
            subprocess.check_call(["kill", "--signal", "SIGINT", str(parent.pid)])

        time.sleep(timeout)

    return ppg.FileGeneratingJob(
        f"all_cores{ii}",
        inner,
        resources=ppg.Resources.AllCores,
        depend_on_function=False,
    )


ac = [all_cores(ii) for ii in range(jobcount)]

start = time.time()
try:
    ppg.run()
except KeyboardInterrupt:
    print("Received expected Keyboard interrupt")
    stop = time.time()
    print(
        f"stop took {stop-start-time_before_abort-1:.2f} seconds.\n "
        "We expected less than{timeout}.\n"
    )



#!/usr/bin/env python3

"""This creates one FileGeneratingJob with a few hundred of direct DataLoadingJobs,
which exposed a o(n**3) or so bottleneck in runner.modify_dag
"""
import pypipegraph2 as ppg
import shutil
from pathlib import Path
import os
import sys

run_dir = (Path(__file__).parent / "run").absolute()

try:
    count = int(sys.argv[1])
except:  # noqa:E722
    count = 200


def simple():
    if run_dir.exists():
        shutil.rmtree(run_dir)
    run_dir.mkdir()
    os.chdir(run_dir)
    ppg.new()

    data = {}

    def final(of):
        of.write_text(str(len(data)))

    final_job = ppg.FileGeneratingJob("final", final)

    last = final_job
    for ii in range(count):

        def load(ii=ii):
            data[ii] = str(ii)

        loader = ppg.DataLoadingJob(f"dl{ii}", load)
        last.depends_on(loader)
        last = loader
    ppg.run()


def test_simple(benchmark):
    print(run_dir)
    benchmark(simple)


if __name__ == "__main__":
    simple()



"""Parlallel computation / threading support helpers"""
from threading import Lock, Condition
import ctypes


class _CoreLockContextManager:
    def __init__(self, core_lock, cores):
        self.core_lock = core_lock
        self.cores = cores

    def __enter__(self):
        self.core_lock._acquire(self.cores)

    def __exit__(self, exc_type, exc_value, traceback):
        self.core_lock._release(self.cores)


class FakeContextManager:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class FakeLock:
    def __init__(self):
        self.cm = FakeContextManager()

    def using(self, _cores):
        return self.cm


class CoreLock:
    """Allow threads to request n 'cores',
    if they're available, let it proceed.
    If they're not available, block.
    If they exceed the maxmimum number available: raise.

    Essentially, this is a Semaphore with multi-resource-one-step-acquisition.

    """

    def __init__(self, max_cores):
        self.max_cores = int(max_cores)
        self.remaining = max_cores
        self.lock = Lock()
        self.condition = Condition()
        self.terminated = False

    def using(self, cores):
        return _CoreLockContextManager(self, cores)

    def _acquire(self, count):
        # logger.info(f" {_thread.get_ident()} - acquire({count}) called")
        if count > self.max_cores:
            raise ValueError(f"Count {count} > max_cores {self.max_cores}")
        if count == 0:
            raise ValueError("Count == 0")
        while True:
            with self.lock:
                if self.remaining >= count:
                    self.remaining -= count
                    # logger.info(f"{_thread.get_ident()}, had available")
                    return

            # not enough remaining. try again once somebody releases something
            with self.condition:
                self.condition.wait()
            # logger.info(f"{_thread.get_ident()} condition triggered")

    def _release(self, count):
        # logger.info(f"{_thread.get_ident()} release({count}) called")
        if count == 0:  # pragma: no cover
            raise ValueError("Count == 0")
        with self.lock:
            self.remaining += count
            if self.remaining > self.max_cores:  # pragma: no cover
                raise ValueError("Remaining exceeded max_cores")

        # logger.info(f"{_thread.get_ident()} remaning: {self.remaining}")
        with self.condition:
            self.condition.notify_all()


def async_raise(target_tid, exception):
    """Raises an asynchronous exception in another thread.
    Read http://docs.python.org/c-api/init.html#PyThreadState_SetAsyncExc
    for further enlightenments.
    :param target_tid: target thread identifier
    :param exception: Exception class to be raised in that thread
    """
    # Ensuring and releasing GIL are useless since we're not in C
    # gil_state = ctypes.pythonapi.PyGILState_Ensure()
    ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(target_tid), ctypes.py_object(exception)
    )
    # ctypes.pythonapi.PyGILState_Release(gil_state)
    if ret == 0:
        raise ValueError("Invalid thread ID {}".format(target_tid))
    elif ret > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(target_tid), None)
        raise SystemError("PyThreadState_SetAsyncExc failed")
